'''
Created on Apr 14, 2020

@author: Vlad
'''

import time
import argparse
import sys
import traceback

from align.aligner import mainAlignmentTask
from configuration import buildConfigs, Configs
from tasks import manager

def main():   
    '''
    Resolve the args/configs, spin up the task manager (which deals with worker threads and handles parallelism), 
    and get started on the main alignment task. 
    '''
    
    startTime = time.time()
    args = parseArgs()
    buildConfigs(args)    
    Configs.log("MAGUS was run with: {}".format(" ".join(sys.argv)))
    
    try:
        manager.startTaskManager()
        mainAlignmentTask()
    except:
        Configs.error("MAGUS aborted with an exception..")
        Configs.error(traceback.format_exc())
    finally:
        manager.stopTaskManager()
    
    endTime = time.time()
    Configs.log("MAGUS finished in {} seconds..".format(endTime-startTime))
    
def parseArgs():
    parser = argparse.ArgumentParser()

    parser.add_argument("-d", "--directory", type=str,
                        help="Path to working directory", required=False, default=None)
    
    parser.add_argument("-i", "--sequences", type=str,
                        help="Path to input unaligned sequences", required=False, default=None)
    
    parser.add_argument("-s", "--subalignments", type=str, nargs="+",
                        help="Paths to input subalignment files", required=False, default=[])
    
    parser.add_argument("-b", "--backbones", type=str, nargs="+",
                        help="Paths to input backbone alignment files", required=False, default=[])

    parser.add_argument("-o", "--output", type=str,
                        help="Output alignment path", required=True)
    
    parser.add_argument("-t", "--guidetree", type=str,
                        help="Guide tree for subset decomposition. fasttree (default), fasttree-noml, clustal, parttree, or path to user guide tree",
                        required=False, default="fasttree")

    parser.add_argument("-np", "--numprocs", type=int,
                        help="Number of processors to use (default: # cpus available)",
                        required=False, default=-1)
    
    parser.add_argument("--maxsubsetsize", type=int,
                        help="Maximum subset size for divide-and-conquer",
                        required=False, default=50)
    
    parser.add_argument("--maxnumsubsets", type=int,
                        help="Maximum number of subsets for divide-and-conquer",
                        required=False, default=25)
    
    parser.add_argument("--decompstrategy", type=str,
                        help="Initial decomposition strategy (pastastyle or kmh)",
                        required=False, default="pastastyle")
    
    parser.add_argument("--decompskeletonsize", type=int,
                        help="Number of skeleton sequences for the initial decomposition strategy",
                        required=False, default=300)
    
    parser.add_argument("--datatype", type=str,
                        help="Data type (dna, rna, or protein). Will be inferred if not provided",
                        required=False, default=None)
    
    parser.add_argument("--graphbuildmethod", type=str,
                        help="Method for building the alignment graph (mafft, mafftmerge, or initial)",
                        required=False, default="mafft")
    
    parser.add_argument("--graphbuildrestrict", type=str,
                        help="Prevent the alignment graph from adding edges that violate subalignments (true or false)",
                        required=False, default="False")
    
    parser.add_argument("--graphbuildhmmextend", type=str,
                        help="Extend the alignment graph MAFFT backbones with hmmer (true or false)",
                        required=False, default="False")
    
    parser.add_argument("--graphclustermethod", type=str,
                        help="Method for initial clustering of the alignment graph (mcl or none)",
                        required=False, default="mcl")
    
    parser.add_argument("--graphtracemethod", type=str,
                        help="Method for finding a trace from the alignment graph (minclusters, fm, mwtgreedy, or mwtsearch)",
                        required=False, default="minclusters")
    
    parser.add_argument("--graphtraceoptimize", type=str,
                        help="Run an optimization step on the graph trace (true or false)",
                        required=False, default="False")
    
    parser.add_argument("-r", "--mafftruns", type=int,
                        help="Number of MAFFT runs", required=False, default=10)
    
    parser.add_argument("-m", "--mafftsize", type=int,
                        help="Maximum size of MAFFT alignments", required=False, default=200)
    
    parser.add_argument("-f", "--inflationfactor", type=float,
                        help="MCL inflation factor", required=False, default=4)
    
    parser.add_argument("-c", "--constrain", type=str,
                        help="Constrain MAGUS to respect subalignments (true or false)", required=False, default="true")
    
    parser.add_argument("--onlyguidetree", type=str,
                        help="Only output the guide tree (true or false)", required=False, default="false")
    
    parser.add_argument("--recurse", type=str,
                        help="Allow MAGUS to recurse on large subsets (true or false)", required=False, default="true")
    
    parser.add_argument("--recurseguidetree", type=str,
                        help="If recursing, passes this argument as the guide tree option to the lower levels. (Default fasttree)", required=False, default="fasttree")
    
    parser.add_argument("--recursethreshold", type=int,
                        help="MAGUS will recursively align subsets above this threshold size", required=False, default=200)
    
    parser.add_argument("--alignsizelimit", type=float,
                        help="Size threshold for alignment compression (in GB)", required=False, default=100)
       
    return parser.parse_args()

if __name__ == '__main__':
    main()


'''
Created on May 28, 2020

@author: Vlad
'''

import re
import os
import math
from tools import external_tools
from configuration import Configs
from helpers import sequenceutils

'''
def buildHmmScores(hmmPath, queriesPath, scoreMap):
    #tasks = [getHmmScores(hmmPath, queriesPath) for hmmPath in hmmPaths]
    queries = sequenceutils.readFromFasta(queriesPath, removeDashes = True)
    baseName = os.path.basename(queriesPath).split('.')[0]
    dirName = os.path.dirname(hmmPath)
    
    taxonScores = {}
    for taxon in queries:
        inputName = os.path.join(dirName, "{}.txt".format(baseName))
        outputName = os.path.join(dirName, "{}_score.txt".format(baseName))
        if os.path.exists(outputName):
            os.remove(outputName)
        sequenceutils.writeFasta({taxon : queries[taxon]}, inputName)
        getHmmScores(hmmPath, inputName, outputName).run()
        subsetScores = readSearchFile(outputName)
        taxonScores[taxon] = subsetScores[taxon][1]
    scoreMap[hmmPath] = taxonScores
'''

def buildHmmScores(hmmPaths, queriesPath, scoreFileHmmFileMap):
    #tasks = [getHmmScores(hmmPath, queriesPath) for hmmPath in hmmPaths]
    queries = sequenceutils.readFromFasta(queriesPath, removeDashes = True)
    baseName = os.path.basename(queriesPath).split('.')[0]
    dirName = os.path.join(os.path.dirname(queriesPath), "chunks_{}".format(baseName))
    if not os.path.exists(dirName):
        os.makedirs(dirName)
    
    chunkSize = 1000
    
    taxa = list(queries.keys())
    inputOutputs = []
    for i in range(math.ceil(len(taxa) / chunkSize)):
        chunk = taxa[i*chunkSize : min(len(taxa), (i+1)*chunkSize)]
        inputName = os.path.join(dirName, "{}_chunk_{}.txt".format(baseName, i+1))
        sequenceutils.writeFasta(queries, inputName, chunk)
        for hmmPath in hmmPaths:
            outputName = os.path.join(os.path.dirname(hmmPath), "{}_chunk_{}_score.txt".format(baseName, i+1))
            inputOutputs.append((hmmPath, inputName, outputName))
            scoreFileHmmFileMap[outputName] = hmmPath
    
    tasks = [getHmmScores(hmmPath, inputPath, outputPath) for hmmPath, inputPath, outputPath in inputOutputs]
    return tasks

def getHmmScores(hmmPath, queriesPath, scorePath):
    workingDir = os.path.dirname(hmmPath)
    #searchPath = os.path.join(workingDir, "hmm_search.txt")
    task = external_tools.runHmmSearch(hmmPath, queriesPath, workingDir, scorePath)
    return task

def readHmmScores(searchFiles):
    sequenceScores = {}
    for file in searchFiles:
        subsetScores = readSearchFile(file)
        taxonScores = {taxon : scores[1] for taxon, scores in subsetScores.items()}
        sequenceScores[file] = taxonScores
    return sequenceScores
    
def buildHmms(sequencesHmmsPathsMap):
    tasks = [buildHmmOverAlignment(sequencePath, hmmPath) for sequencePath, hmmPath in sequencesHmmsPathsMap.items()]
    return tasks
    
def buildHmmOverAlignment(sequencePath, hmmPath):
    workingDir = os.path.dirname(hmmPath)
    task = external_tools.runHmmBuild(sequencePath, workingDir, hmmPath)
    return task

def combineHmmAlignments(alignFiles, outputAlignmentPath, includeInsertions):
    alignment = {}
    for file in alignFiles:
        alignment.update(sequenceutils.readFromStockholm(file, includeInsertions))
    sequenceutils.writeFasta(alignment, outputAlignmentPath, None)

def mergeHmmAlignments(alignFiles, outputAlignmentPath, includeInsertions):
    for file in alignFiles:
        alignment = sequenceutils.readFromStockholm(file, includeInsertions)
        sequenceutils.writeFasta(alignment, outputAlignmentPath, None, True)

def hmmAlignQueries(hmmPath, queriesPath):
    queries = sequenceutils.readFromFasta(queriesPath, removeDashes = True)
    baseName = os.path.basename(queriesPath).split('.')[0]
    dirName = os.path.join(os.path.dirname(queriesPath), "chunks_{}".format(baseName))
    if not os.path.exists(dirName):
        os.makedirs(dirName)
    chunkSize = 1000
    
    taxa = list(queries.keys())
    alignFiles = {}
    for i in range(math.ceil(len(taxa) / chunkSize)):
        chunk = taxa[i*chunkSize : min(len(taxa), (i+1)*chunkSize)]
        inputName = os.path.join(dirName, "{}_chunk_{}.txt".format(baseName, i+1))
        outputName = os.path.join(dirName, "{}_chunk_{}_aligned.txt".format(baseName, i+1))
        sequenceutils.writeFasta(queries, inputName, chunk)
        alignFiles[inputName] = outputName
    
    tasks = []
    for inputPath, outputPath in alignFiles.items():
        task = buildHmmAlignment(hmmPath, inputPath, outputPath)
        tasks.append(task)
    return tasks

def buildHmmAlignment(hmmPath, queriesPath, outputAlignmentPath):
    workingDir = os.path.dirname(hmmPath)
    task = external_tools.runHmmAlign(hmmPath, queriesPath, workingDir, outputAlignmentPath)
    return task

#from PASTA repo
def readSearchFile(searchFilePath):
    with open(searchFilePath, 'r') as searchFile:
        results = {}

        pattern = re.compile(
            r"([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+"
            r"([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)")
        start_reading = False
        for line in searchFile:
            line = line.strip()
            if (not start_reading and line.startswith("E-value") is True):
                start_reading = True
            elif (start_reading and line == ""):
                start_reading = False
                break
            elif (start_reading):
                matches = pattern.search(line)
                if (matches is not None and matches.group(0).find("--") == -1):
                    results[matches.group(9).strip()] = (
                        float(matches.group(1).strip()),
                        float(matches.group(2).strip()))
                    # _LOG.debug("Fragment scores;"
                    #           "fragment:%s E-Value:%s BitScore:%s" %(matches
                    # .group(9).strip(),matches.group(1).strip(), matches.
                    # group(2).strip()))
        return results


'''
Created on Oct 26, 2020

@author: Vlad
'''

import random
import threading
import os
import time
import concurrent.futures
from configuration import Configs
from tasks import files

'''
Launching and awaiting tasks.
To avoid deadlocks and stack overflows, only the main thread can submit tasks.
Thus, only the main thread runs alignment tasks, worker threads are used for other task types (like MAFFT).
'''

class TaskManager():
    
    runningTasksFile = None
    lockTasksFile = None
    
    managerPool = None
    managerFuture = None
    managerSignal = threading.Event()
    managerLock = threading.Lock()
    managerStopSignal = False
    
    observerSignal = threading.Event()
    observerWaiting = False
    observerTask = None 
    
    waitingTasks = {}
    submittedTasks = set()
    runningTasks = set()
    finishedTasks = set()
    failedTasks = set()
    
    taskPool = None
    threadsUsed = 0
    lastFilesCheckTime = 0
    lastDebugTime = 0
    serialTaskTypes = {"runAlignmentTask", "buildInducedSubalignment", "compressSubalignment"}
    contextStack = []
    
def startTaskManager():
    Configs.debug("Starting up the task manager..")
    
    tasksDir = os.path.join(Configs.workingDir, "tasks")
    TaskManager.pendingTasksDir = os.path.join(tasksDir, "tasks_pending")
    TaskManager.runningTasksFile = os.path.join(tasksDir, "tasks_running.txt")
    TaskManager.lockTasksFile = os.path.join(tasksDir, "tasks.lock")
    if not os.path.exists(TaskManager.pendingTasksDir):
        os.makedirs(TaskManager.pendingTasksDir)  
    
    TaskManager.managerPool = concurrent.futures.ThreadPoolExecutor(max_workers = 1)
    TaskManager.managerFuture = TaskManager.managerPool.submit(runTaskManager)
    TaskManager.taskPool = concurrent.futures.ThreadPoolExecutor(max_workers = Configs.numCores)
    Configs.debug("Task manager is up..")

def stopTaskManager():
    TaskManager.managerStopSignal = True
    with TaskManager.managerLock:
        TaskManager.managerSignal.set()
    try:
        Configs.debug("Winding down the task manager..")
        TaskManager.managerFuture.result()
    finally:
        Configs.log("Waiting for {} tasks to finish..".format(len(TaskManager.runningTasks)))
        TaskManager.taskPool.shutdown()
        dealWithFinishedTasks()        
        TaskManager.managerPool.shutdown()
        Configs.debug("Task manager stopped..")

def runTaskManager():
    try:
        while not TaskManager.managerStopSignal:
            with TaskManager.managerLock:
                dealWithErrors()
                dealWithFinishedTasks()
                dealWithPendingTasks()
                dealWithWaitingTasks()
                TaskManager.managerSignal.clear()
            TaskManager.managerSignal.wait(5)
    finally:
        TaskManager.observerSignal.set()  

def dealWithErrors():
    for task in TaskManager.failedTasks:
        Configs.error("Task manager found a failed task: {}".format(task.outputFile))
        if task.future is not None:
            task.future.result()
        
def dealWithFinishedTasks():
    stoppedRunning = set(t for t in TaskManager.finishedTasks | TaskManager.failedTasks if t.taskType == "runAlignmentTask")
    if len(stoppedRunning) > 0:
        with files.FileLock(TaskManager.lockTasksFile):
            runningTasks = files.readTasksFromFile(TaskManager.runningTasksFile)
            stillRunningTasks = [t for t in runningTasks if t not in stoppedRunning]
            if len(stillRunningTasks) < len(runningTasks):
                files.writeTasksToFile(stillRunningTasks, TaskManager.runningTasksFile, append = False)
        
    if len(TaskManager.failedTasks) > 0:
        processPendingTasks(TaskManager.failedTasks, 0, None)   
        
    TaskManager.finishedTasks = set()
    TaskManager.failedTasks = set()

def dealWithPendingTasks():
    numToLaunch = min(1, Configs.numCores - TaskManager.threadsUsed)
    newTasks = []
    for t in TaskManager.submittedTasks:
        if os.path.exists(t.outputFile):
            Configs.log("File already exists: {}".format(t.outputFile))
            t.isFinished = True
            TaskManager.observerSignal.set()
        else:
            newTasks.append(t)
            TaskManager.waitingTasks[t.outputFile] = t
    
    if len(newTasks) > 0:
        launchedTasks, remainingTasks = processPendingTasks(newTasks, numToLaunch, None)
        numToLaunch = numToLaunch - len(launchedTasks)
    if numToLaunch > 0:
        pendingFiles = [os.path.join(TaskManager.pendingTasksDir, file) for file in os.listdir(TaskManager.pendingTasksDir) if file.endswith(".txt")]
        random.shuffle(pendingFiles)
        for taskFile in pendingFiles:
            if numToLaunch <= 0:
                break
            launchedTasks, remainingTasks = processPendingTasks(None, numToLaunch, taskFile)
            numToLaunch = numToLaunch - len(launchedTasks)

    TaskManager.submittedTasks = set()

def dealWithWaitingTasks():
    timeSinceFileCheck = time.time() - TaskManager.lastFilesCheckTime
    if timeSinceFileCheck >= 5: 
        for file, task in list(TaskManager.waitingTasks.items()):
            if os.path.exists(file):
                Configs.debug("Detected task completion: {}".format(file))
                TaskManager.waitingTasks.pop(file)
                task.isFinished = True
                TaskManager.observerSignal.set()
        TaskManager.lastFilesCheckTime = time.time()
    
    timeSinceDebug = time.time() - TaskManager.lastDebugTime
    if timeSinceDebug >= 60:  
        TaskManager.lastDebugTime = time.time()
        for task in TaskManager.runningTasks:
            Configs.debug("Still running task {}, status {}".format(task.outputFile, task.future._state if task.future is not None else "N/A"))
        for file in TaskManager.waitingTasks:
            Configs.debug("Still waiting on task {}".format(file))

def processPendingTasks(tasks, numTasksToLaunch, taskFile):
    if taskFile is None:
        taskFile = os.path.join(TaskManager.pendingTasksDir, "task_file_{:03d}.txt".format(random.randint(0,999)))
    
    newTasks = True    
    with files.FileLock(taskFile.replace(".txt", ".lock")):
        if tasks is None:
            tasks = files.readTasksFromFile(taskFile)
            newTasks = False

        alignTasks = [t for t in tasks if t.taskType == "runAlignmentTask"]
        if len(alignTasks) > 0:
            with files.FileLock(TaskManager.lockTasksFile):
                runningTasks = set(files.readTasksFromFile(TaskManager.runningTasksFile))
                if len(runningTasks) > 0:
                    tasks = [t for t in tasks if t not in runningTasks]
                launchedTasks, remainingTasks = launchTasks(tasks, numTasksToLaunch)        
                writeRunningTasks = [t for t in launchedTasks if t.taskType == "runAlignmentTask"]
                files.writeTasksToFile(writeRunningTasks, TaskManager.runningTasksFile, append = True)
        else:
            launchedTasks, remainingTasks = launchTasks(tasks, numTasksToLaunch)
        
        files.writeTasksToFile(remainingTasks, taskFile, append = newTasks) 
        if len(launchedTasks) > 0:     
            Configs.debug("Launched {} tasks and deferred {} tasks..".format(len(launchedTasks), len(remainingTasks)))   
          
    return launchedTasks, remainingTasks      

def launchTasks(tasks, numTasksToLaunch):
    launchedTasks = []
    remainingTasks = []    
    threadsAvailable = Configs.numCores - TaskManager.threadsUsed
    toLaunch = min(numTasksToLaunch, threadsAvailable)    
    for task in tasks:
        if toLaunch > 0 and checkLaunchTask(task):
            Configs.debug("Launched a new task.. {}/{} threads used, type: {}, output file: {}".format(TaskManager.threadsUsed, Configs.numCores, task.taskType, task.outputFile))
            toLaunch = toLaunch - 1
            launchedTasks.append(task)
        else:
            remainingTasks.append(task)
    return launchedTasks, remainingTasks

def checkLaunchTask(task):
    if task.taskType not in TaskManager.serialTaskTypes:
        task.future = TaskManager.taskPool.submit(runTask, task)
        return True
    elif TaskManager.observerWaiting and TaskManager.observerTask is None:
        stack = TaskManager.contextStack
        if task.taskType != "runAlignmentTask" or len(stack) == 0 or task in stack[-1].subalignmentTasks:
            TaskManager.observerTask = task
            TaskManager.observerSignal.set()
            return True
    return False

def runTask(task):
    with TaskManager.managerLock:
        if task.taskType not in TaskManager.serialTaskTypes:
            TaskManager.threadsUsed = TaskManager.threadsUsed + 1
        TaskManager.runningTasks.add(task)
    
    failed = False
    try:
        task.run()     
    except:
        failed = True
        raise
    finally:
        with TaskManager.managerLock:
            if task.taskType not in TaskManager.serialTaskTypes:
                TaskManager.threadsUsed = TaskManager.threadsUsed - 1
            TaskManager.runningTasks.remove(task)
            TaskManager.failedTasks.add(task) if failed else TaskManager.finishedTasks.add(task)
            if task.outputFile in TaskManager.waitingTasks:
                t = TaskManager.waitingTasks.pop(task.outputFile)
                t.future = task.future
                t.isFinished = True
            TaskManager.managerSignal.set()
            TaskManager.observerSignal.set()
                
        


'''
Created on Aug 23, 2020

@author: Vlad
'''

import os
import time

from configuration import Configs

from align.merge.graph_cluster.mcl import runMclClustering
from align.merge.graph_cluster.mlr_mcl import runMlrMclClustering
from align.merge.graph_cluster.rg import rgClustering

'''
The alignment graph is clustered, the clusters are written out as an array of node arrays.
MCL is the main way to do this, but rg could be used if there are scalability issues.
'''

def clusterGraph(graph):
    time1 = time.time()
    
    if os.path.exists(graph.clusterPath):
        Configs.log("Found existing cluster file {}".format(graph.clusterPath))
        graph.readClustersFromFile(graph.clusterPath)
        
    elif Configs.graphClusterMethod == "mcl":
        runMclClustering(graph)      
        
    elif Configs.graphClusterMethod == "mlrmcl":
        runMlrMclClustering(graph)
        
    elif Configs.graphClusterMethod == "rg":
        rgClustering(graph)
        
    else:
        Configs.log("No alignment graph clustering requested..")
    
    time2 = time.time()  
    Configs.log("Clustered the graph in {} sec..".format(time2-time1))



'''
Created on Apr 14, 2020

@author: Vlad
'''

import os
import time
import random

from align.merge.alignment_graph import AlignmentGraph
from helpers import sequenceutils, hmmutils
from tasks import task
from configuration import Configs
from tools import external_tools

'''
Building a MAGUS alignment graph from backbone alignments. 
The graph is a sparse matrix, stored as a weighted adjacency list.
Backbone alignment tasks are run in parallel, and can begin before the subalignments are finished.
When subalignments finish, we can initialize the alignment graph. 
Backbones are added to this graph as they complete.
Graph is written to file in MCL-compliant format.
'''

def buildGraph(context):
    time1 = time.time() 
    
    context.graph = AlignmentGraph(context)
    context.initializeSequences()
    
    if os.path.exists(context.graph.graphPath):
        Configs.log("Found existing graph file {}".format(context.graph.graphPath))
    else:
        requestBackboneTasks(context)
    
    context.awaitSubalignments()
    context.graph.initializeMatrix()
    
    if os.path.exists(context.graph.graphPath):
        context.graph.readGraphFromFile(context.graph.graphPath)
    else:
        context.initializeBackboneSequenceMapping()
        buildMatrix(context)
        context.graph.writeGraphToFile(context.graph.graphPath)
       
    time2 = time.time()
    Configs.log("Built the alignment graph in {} sec..".format(time2-time1))

def requestBackboneTasks(context):
    if len(context.backbonePaths) > 0:
        Configs.log("Using {} user-defined backbone files..".format(len(context.backbonePaths)))
        context.backbonePaths = context.backbonePaths
        for path in context.backbonePaths:
            context.backboneTaxa.update(sequenceutils.readFromFasta(path))
    
    elif Configs.graphBuildMethod == "mafft":
        Configs.log("Using {} MAFFT backbones..".format(Configs.mafftRuns)) 
        requestMafftBackbones(context)
        
    elif Configs.graphBuildMethod == "subsethmm":
        Configs.log("Using {} HMM-extended subalignments as backbone files..".format(len(context.subalignmentPaths)))
        context.backbonePaths = context.subalignmentPaths
        context.backboneExtend.update(context.backbonePaths)
        
    elif Configs.graphBuildMethod == "initial":
        Configs.log("Using the initial decomposition alignment as the single backbone..")
        initialAlignPath = os.path.join(context.workingDir, "decomposition", "initial_tree", "initial_insert_align.txt")
        context.backbonePaths = [initialAlignPath]
    
    if not Configs.constrain and Configs.graphBuildMethod != "subsethmm":
        context.backbonePaths.extend(context.subalignmentPaths)

def requestMafftBackbones(context):
    numTaxa = max(1, int(Configs.mafftSize/len(context.subsetPaths)))

    for n in range(Configs.mafftRuns):
        unalignedFile = os.path.join(context.graph.workingDir, "backbone_{}_unalign.txt".format(n+1))
        alignedFile = os.path.join(context.graph.workingDir, "backbone_{}_mafft.txt".format(n+1))
        if os.path.exists(alignedFile):
            Configs.log("Existing backbone file found: {}".format(alignedFile))
            backbone = sequenceutils.readFromFasta(alignedFile)
            context.backbonePaths.append(alignedFile)
        else:
            backbone = assignBackboneTaxa(context, numTaxa, unalignedFile)
            backboneTask = external_tools.buildMafftAlignment(unalignedFile, alignedFile)
            context.backboneTasks.append(backboneTask)
            
        if Configs.graphBuildHmmExtend:
            context.backboneExtend.add(alignedFile)
        else:
            context.backboneTaxa.update(backbone)
    task.submitTasks(context.backboneTasks)    
                
def buildMatrix(context):
    addedBackbones = set()
    for backboneTask in task.asCompleted(context.backboneTasks):
        addAlignmentFileToGraph(context, backboneTask.outputFile)
        addedBackbones.add(backboneTask.outputFile)
    
    for backboneFile in context.backbonePaths:
        if backboneFile not in addedBackbones:
            addAlignmentFileToGraph(context, backboneFile)  
    
def assignBackboneTaxa(context, numTaxa, unalignedFile):
    backbone = {}
    for subset in context.subsets:
        random.shuffle(subset)
        for taxon in subset[:numTaxa]:
            backbone[taxon] = context.unalignedSequences[taxon]
    sequenceutils.writeFasta(backbone, unalignedFile)
    return backbone
    
def addAlignmentFileToGraph(context, alignedFile):
    Configs.log("Feeding backbone {} to the graph..".format(alignedFile))
    backboneAlign = sequenceutils.readFromFasta(alignedFile)  
    alignmentLength = len(next(iter(backboneAlign.values())).seq)
     
    if alignedFile in context.backboneExtend:
        extensionTasks = requestHmmExtensionTasks(context, backboneAlign, alignedFile)
        task.submitTasks(extensionTasks)
        for extensionTask in task.asCompleted(extensionTasks):
            backboneAlign.update(sequenceutils.readFromStockholm(extensionTask.outputFile, includeInsertions=True))
    
    alignmap = backboneToAlignMap(context, backboneAlign, alignmentLength)
    Configs.log("Constructed backbone alignment map from {}".format(alignedFile))
    
    graph = context.graph  
    with graph.matrixLock:
        for l in range(alignmentLength):  
            for a, avalue in alignmap[l].items():
                for b, bvalue in alignmap[l].items():
                    
                    if Configs.graphBuildRestrict:
                        asub, apos = graph.matSubPosMap[a] 
                        bsub, bpos = graph.matSubPosMap[b] 
                        if asub == bsub and apos != bpos:
                            continue
                                
                    graph.matrix[a][b] = graph.matrix[a].get(b,0) + avalue * bvalue         
    Configs.log("Fed backbone {} to the graph.".format(alignedFile))

def backboneToAlignMap(context, backboneAlign, alignmentLength):
    alignmap = [{} for i in range(alignmentLength)]
    t = 0
    
    for taxon in backboneAlign:
        subsetIdx = context.taxonSubalignmentMap[taxon]
        subsetseq = context.backboneSubalignment[taxon].seq
        unalignedseq = context.unalignedSequences[taxon].seq     
        backboneseq = backboneAlign[taxon].seq   
        
        i = 0
        posarray = [0] * len(unalignedseq)
        for n in range(len(subsetseq)):
            if subsetseq[n] == unalignedseq[i]:
                posarray[i] = n 
                i = i + 1
                if i == len(unalignedseq):
                    break
        
        i = 0
        n = 0
        for c in backboneseq:
            if i == len(unalignedseq):
                break
            if c == unalignedseq[i]:
                position = int(context.graph.subsetMatrixIdx[subsetIdx] + posarray[i])
                alignmap[n][position] = alignmap[n].get(position, 0) + 1
            if c.upper() == unalignedseq[i]:
                i = i + 1
            if c == c.upper() and c != '.':
                n = n + 1
                
        t = t + 1
            
    return alignmap
        
def requestHmmExtensionTasks(context, backbone, alignedFile):
    baseName = os.path.basename(alignedFile)
    hmmDir = os.path.join(context.graph.workingDir, "hmm_{}".format(baseName))
    extensionUnalignedFile = os.path.join(hmmDir, "queries.txt")
    hmmPath = os.path.join(hmmDir, "hmm_model.txt")
    if not os.path.exists(hmmDir):
        os.makedirs(hmmDir)
    
    backboneExtension = {}
    for taxon in context.unalignedSequences:
        if not taxon in backbone:
            backboneExtension[taxon] = context.unalignedSequences[taxon]
                
    sequenceutils.writeFasta(backboneExtension, extensionUnalignedFile) 
    buildTask = hmmutils.buildHmmOverAlignment(alignedFile, hmmPath)
    buildTask.run()
    alignTasks = hmmutils.hmmAlignQueries(hmmPath, extensionUnalignedFile)
    return alignTasks 
    


'''
Created on Dec 4, 2020

@author: Vlad
'''

import os
from helpers import sequenceutils
from tasks import task, manager
from configuration import Configs

'''
The AlignmentContext data structure maintains all information pertaining to a single MAGUS alignment.
The main thread keeps one context active at a time (to manage resources). Subalignments will spawn their
own contexts, which will become active when the parent context is in a blocking wait. The previously active
context resumes when the current alignment is completed.
'''

class AlignmentContext:
        
    def __init__(self, **kwargs):
        self.outputFile = None
        self.workingDir = None
        self.sequencesPath = None
        self.subsetPaths = []
        self.subalignmentPaths = []
        self.backbonePaths = []
        self.guideTree = None
        
        self.unalignedSequences = None
        #self.taxa = []
        self.subsets = []
        self.subalignments = []
        self.taxonSubsetMap = {}
        self.taxonSubalignmentMap = {}
                
        self.backboneTaxa = {}
        self.backboneExtend = set()
        self.backboneSubalignment = {}
        
        self.subalignmentTasks = []
        self.backboneTasks = []
        self.graph = None
        
        for attr in kwargs:
            vars(self)[attr] = kwargs.get(attr)
        
        if not os.path.exists(self.workingDir):
            os.makedirs(self.workingDir)
    
    def awaitSubalignments(self):
        task.awaitTasks(self.subalignmentTasks)
    
    def initializeSequences(self):
        self.unalignedSequences = {}
        for i, subsetPath in enumerate(self.subsetPaths):
            self.subsets.append([])
            subset = sequenceutils.readFromFastaOrdered(subsetPath, removeDashes=True)
            for sequence in subset:
                self.unalignedSequences[sequence.tag] = sequence
                self.taxonSubsetMap[sequence.tag] = i
                self.subsets[i].append(sequence.tag)
        
        if Configs.constrain:
            self.subalignments = self.subsets
            self.taxonSubalignmentMap = self.taxonSubsetMap
        else:
            for s in self.subsets:
                for taxon in s:
                    self.taxonSubalignmentMap[taxon] = len(self.subalignments)
                    self.subalignments.append([taxon])
                
    def initializeBackboneSequenceMapping(self):
        if len(self.backboneTaxa) == 0:
            backboneSubsetTaxonMap = {i : subset for i, subset in enumerate(self.subsets)}
        else:
            backboneSubsetTaxonMap = {}
            for taxon in self.backboneTaxa:
                i = self.taxonSubsetMap[taxon]
                backboneSubsetTaxonMap[i] = backboneSubsetTaxonMap.get(i, [])
                backboneSubsetTaxonMap[i].append(taxon) 
        
        if Configs.constrain:
            for i, subalignPath in enumerate(self.subalignmentPaths):
                subalignment = sequenceutils.readFromFasta(subalignPath, removeDashes=False)
                for taxon in backboneSubsetTaxonMap.get(i, []):
                    self.backboneSubalignment[taxon] = subalignment[taxon]
        else:
            self.backboneSubalignment = self.unalignedSequences
    
    def __enter__(self):
        manager.TaskManager.contextStack.append(self)
        return self
            
    def __exit__(self, excType, excVal, excTb):
        manager.TaskManager.contextStack.pop()
    


'''
Created on Jul 17, 2020

@author: Vlad
'''

import os
import time

from configuration import Configs
from align.merge.graph_cluster.clean_clusters import purgeClusterViolations, purgeDuplicateClusters
from align.merge.graph_trace.min_clusters import minClustersSearch
from align.merge.graph_trace.fm import fmAlgorithm
from align.merge.graph_trace.mwt_search import mwtGreedySearch, mwtSearch
from align.merge.graph_trace.rg_search import rgSearch
from align.merge.graph_trace.rg_fast_search import rgFastSearch
from align.merge.graph_trace.naive import naiveClustering

'''
Graph clusters must be refined into a "trace", a constrained clustering that corresponds to a valid MSA.
There are a variety of ways to do this, "minclusters" is usually the most dependable option.
"mwtgreedy" or "rgfast" might be used if there are scalability issues.
Some of these options don't require an existing clustering, and can work on the raw graph.
'''

def findTrace(graph):
    time1 = time.time() 
    
    if os.path.exists(graph.tracePath):
        Configs.log("Found existing trace file {}".format(graph.tracePath))
        graph.readClustersFromFile(graph.tracePath)
        
    else:
        purgeDuplicateClusters(graph)
        purgeClusterViolations(graph)
        
        if Configs.graphTraceMethod == "minclusters":
            minClustersSearch(graph)       
        elif Configs.graphTraceMethod == "fm":            
            fmAlgorithm(graph)        
        elif Configs.graphTraceMethod == "mwtgreedy":
            mwtGreedySearch(graph)
        elif Configs.graphTraceMethod == "mwtsearch":
            mwtSearch(graph)
        elif Configs.graphTraceMethod == "rg":
            rgSearch(graph)
        elif Configs.graphTraceMethod == "rgfast":
            rgFastSearch(graph)
        elif Configs.graphTraceMethod == "naive":
            naiveClustering(graph)
        
        graph.writeClustersToFile(graph.tracePath)
    
    
    time2 = time.time()
    Configs.log("Found alignment graph trace in {} sec..".format(time2-time1))
    Configs.log("Found a trace with {} clusters and a total cost of {}".format(len(graph.clusters), graph.computeClusteringCost(graph.clusters)))
    
    
    
    


'''
Created on Aug 23, 2020

@author: Vlad
'''

from configuration import Configs
from align.merge.graph_trace.rg_search import rgCluster
from align.merge.graph_trace.rg_fast_search import rgFastCluster

def rgClustering(graph):
    Configs.log("Building a region-growing graph clustering..")
    
    k = len(graph.context.subalignments)
    lowerBound = [graph.subsetMatrixIdx[i] for i in range(k)]
    upperBound = [graph.subsetMatrixIdx[i] + graph.subalignmentLengths[i] for i in range(k)] 
    graph.clusters =  rgCluster(graph, lowerBound, upperBound, False)
    graph.writeClustersToFile(graph.clusterPath)

def rgFastClustering(graph):
    Configs.log("Building a fast region-growing graph clustering..")
    
    k = len(graph.context.subalignments)
    lowerBound = [graph.subsetMatrixIdx[i] for i in range(k)]
    upperBound = [graph.subsetMatrixIdx[i] + graph.subalignmentLengths[i] for i in range(k)] 
    graph.clusters =  rgFastCluster(graph, lowerBound, upperBound, False)
    graph.writeClustersToFile(graph.clusterPath)



'''
Created on Jun 25, 2020

@author: Vlad
'''

import heapq
from collections import deque 

from configuration import Configs


'''
Resolve clusters into a trace by looking for cycles and removing edges to break the cycles.
We're done when there are no more cycles.
'''

def mwtGreedySearch(graph):
    Configs.log("Finding graph trace with MWT greedy search..")
    
    k = len(graph.context.subalignments)
    lowerBound = [graph.subsetMatrixIdx[i] for i in range(k)]
    upperBound = [graph.subsetMatrixIdx[i] + graph.subalignmentLengths[i] for i in range(k)] 
    
    if graph.clusters is None or len(graph.clusters) == 0:
        graph.buildNodeEdgeDataStructure()
    else:
        graph.buildNodeEdgeDataStructureFromClusters()
    
    context = MwtSearchContext(lowerBound, upperBound)
    state = MwtSearchState()
    state.frontier = list(lowerBound)    
    clusters, totalCost, cycles = greedySearch(graph, state, context)
    graph.clusters = clusters

def mwtSearch(graph):
    Configs.log("Finding graph trace with MWT heuristic search..")
    
    k = len(graph.context.subalignments)
    lowerBound = [graph.subsetMatrixIdx[i] for i in range(k)]
    upperBound = [graph.subsetMatrixIdx[i] + graph.subalignmentLengths[i] for i in range(k)] 
    
    if graph.clusters is None or len(graph.clusters) == 0:
        graph.buildNodeEdgeDataStructure()
    else:
        graph.buildNodeEdgeDataStructureFromClusters()
    
    clusters, totalCost = mwtHeuristicSearch(graph, lowerBound, upperBound)
    graph.clusters = clusters
    
def mwtHeuristicSearch(graph, lowerBound, upperBound):  
    context = MwtSearchContext(lowerBound, upperBound)
    startState = MwtSearchState()
    startState.frontier = list(lowerBound)
    
    heap = []
    visited = set()    
    maxFrontierState = startState
    heapq.heappush(heap, (startState.getHeuristic(), startState))
    
    while len(heap) > 0:
        
        if len(heap) > Configs.searchHeapLimit:
            Configs.log("Heap limit exceeded, clearing heap and moving to max frontier..")
            heap = []
            visited = set()
            heapq.heappush(heap, (maxFrontierState.getHeuristic(), maxFrontierState))
                
        heuristic, state = heapq.heappop(heap)
 
        newMax = False
        newFull = True
        for i in range(len(state.frontier)): 
            if state.frontier[i] > context.maxFrontier[i]:
                newMax = True
            elif state.frontier[i] == context.maxFrontier[i]:
                newFull = False
            elif state.frontier[i] < context.maxFrontier[i]:
                newMax = False
                newFull = False
                break
        if newMax:
            context.maxFrontier = list(state.frontier)
            maxFrontierState = state
            #Configs.log("New frontier {}..".format(graph_partition.cutString(graph, context.maxFrontier)))
            #Configs.log("{} cost, {} removed..".format(state.cost, len(state.removed)))
        if newFull:
            context.fullFrontier = list(state.frontier)
            #Configs.log("New full frontier {}..".format(graph.cutString(context.fullFrontier)))
            heap = []
            visited = set()
         
        percent = getBoundPercent(state.frontier, context)
        if int(percent/10) > int(context.percentDone/10):
            context.percentDone = percent
            Configs.log("{}% done, {} cost, {} edges removed..".format(percent, state.cost, len(state.removed)))
            Configs.log("Max frontier {}..".format(graph.cutString(context.maxFrontier)))
        
        moves = findMoves(graph, state, context, 1)
        if len(moves) == 0:  
            state.frontier = list(lowerBound)          
            isCycle, orderedClusters = findCycleOrCluster(graph, state, context)
            return orderedClusters, state.cost
        
        for nextState in moves:
            stateKey = nextState.getStateKey()
            if stateKey not in visited:
                visited.add(stateKey)
                heapq.heappush(heap, (nextState.getHeuristic(), nextState))        


    Configs.log("Heap empty, resorting to greedy search..")
    context = MwtSearchContext(lowerBound, upperBound)
    state = MwtSearchState()
    state.frontier = list(lowerBound)
    clusters, totalCost, cycles = greedySearch(graph, state, context)
    return clusters, totalCost

def findMoves(graph, state, context, frontierSearchDepth = 3):
    if len(state.frontierRemoved) >= frontierSearchDepth:
        findGreedyProgress(graph, state, context)
    
    oldFrontier = tuple(state.frontier)    
    isCycle, result = findCycleOrCluster(graph, state, context)
                
    if not isCycle:
        return []
    else:
        moves = []
        sameFrontier = oldFrontier == tuple(state.frontier)
        for edge in result:
            newState = MwtSearchState(state)
            newState.removed.add(edge)
            newState.cost = newState.cost + graph.matrix[edge[0]][edge[1]]
            
            if sameFrontier:
                newState.frontierRemoved.add(edge)
            else:
                newState.frontierRemoved = set([edge])
            moves.append(newState)
        return moves

def findGreedyProgress(graph, state, context):
    state.frontierRemoved = set()
    oldFrontier = tuple(state.frontier)
    
    while True:
        isCycle, result = findCycleOrCluster(graph, state, context)

        if oldFrontier != tuple(state.frontier):
            return
            
        if not isCycle:
            return

        edge = min(result, key = lambda x: graph.matrix[x[0]][x[1]])
        state.removed.add(edge)
        state.cost = state.cost + graph.matrix[edge[0]][edge[1]]

def greedySearch(graph, state, context):
    cycles = []
    isCycle, result = findCycleOrCluster(graph, state, context)
    while isCycle:
        cycles.append(result)
        
        percent = getBoundPercent(state.frontier, context)
        if int(percent/10) > int(context.percentDone/10):
            context.maxFrontier = list(state.frontier)
            context.percentDone = percent
            Configs.log("{}% done, {} cost, {} edges removed..".format(percent, state.cost, len(state.removed)))
            Configs.log("Frontier {}..".format(graph.cutString(context.maxFrontier)))
            
        edge = min(result, key = lambda x: graph.matrix[x[0]][x[1]])
        state.removed.add(edge)
        state.cost = state.cost + graph.matrix[edge[0]][edge[1]]
        
        isCycle, result = findCycleOrCluster(graph, state, context)
        
    state.frontier = list(context.lowerBound)
    isCycle, result = findCycleOrCluster(graph, state, context)
    return result, state.cost, cycles
    
def findCycleOrCluster(graph, state, context):    
    k = len(graph.context.subalignments)
    
    curCluster = None
    clusters = []
    orderedClusters = []
    backPointers = {}
    nodeClusters = {}
    
    while True:
        lowerNode, upperNode = None, None
        if curCluster is None:
            for i in range(k):
                if state.frontier[i] < context.upperBound[i]:
                    lowerNode = state.frontier[i]
                    break
            if lowerNode is None:
                return False, orderedClusters
        else:
            for node in clusters[curCluster]:
                asub, apos = graph.matSubPosMap[node]
                if node > state.frontier[asub]:
                    lowerNode = state.frontier[asub]
                    upperNode = node
                    
                    if lowerNode in nodeClusters:
                        #print("Complex cycle..")
                        other = nodeClusters[lowerNode]
                        path = []                        
                        cur = curCluster
                        startNode = upperNode
                        while cur != other:
                            prev, lower, upper = backPointers[cur]
                            segment = findPathBFS(graph, state.frontier, context.upperBound, state.removed, startNode, lower)
                            path.extend(segment)
                            cur = prev
                            startNode = upper
                        segment = findPathBFS(graph, state.frontier, context.upperBound, state.removed, startNode, lowerNode)
                        return True, path
            
            if lowerNode is None:
                for node in clusters[curCluster]:
                    asub, apos = graph.matSubPosMap[node]
                    state.frontier[asub] = state.frontier[asub] + 1
                clusters[curCluster].sort()
                orderedClusters.append(clusters[curCluster])
                prev, lower, upper = backPointers.get(curCluster, (None, 0, 0))
                curCluster = prev
                continue
            
        isCycle, result = findCycleOrClusterFromNode(graph, state, context, lowerNode)
        if isCycle:
            return True, result
        else:
            clusters.append(result)
            idx = len(clusters)-1    
            for node in result:
                nodeClusters[node] = idx
            backPointers[idx] = (curCluster, lowerNode, upperNode) 
            curCluster = idx                                             
    return False, orderedClusters
    
def findCycleOrClusterFromNode(graph, state, context, node):
    k = len(graph.context.subalignments)
    i, pos = graph.matSubPosMap[node]

    stack = [node]
    clusterNodes = set([node])
    seqPositions = {i : node}
    backPointers = {}
    
    while len(stack) > 0:
        curNode = stack.pop()
        for j in range(k):
            sibling = None
            for nbr, value in graph.nodeEdges[curNode][j]:
                if nbr in clusterNodes or edge(curNode, nbr) in state.removed or nbr < state.frontier[j] or nbr >= context.upperBound[j]:
                    continue
                if sibling is not None:
                    return True, [edge(sibling, curNode), edge(curNode, nbr)]
                sibling = nbr
                
                clusterNodes.add(nbr)
                backPointers[nbr] = curNode
                if j in seqPositions:
                    path = set()
                    cur = nbr
                    while cur in backPointers:
                        prev = backPointers[cur]
                        path.add(edge(cur, prev))
                        cur = prev
                    cur = seqPositions[j]
                    while cur in backPointers:
                        prev = backPointers[cur]
                        e = edge(cur, prev)
                        if e in path:
                            path.remove(e)
                        else:
                            path.add(e)
                        cur = prev
        
                    return True, list(path)           
                
                seqPositions[j] = nbr
                stack.append(nbr)  
    
    return False, list(clusterNodes)            

def findPathBFS(graph, lowerBound, upperBound, removed, nodeA, nodeB):
    k = len(graph.context.subalignments)
    asub, apos = graph.matSubPosMap[nodeA]
    bsub, bpos = graph.matSubPosMap[nodeB]
    #queue = deque([nodeA])
    queue = deque([(nodeA, set([asub]))])
    visited = set([nodeA])
    backPointers = {}
    
    while len(queue) > 0:
        curNode, levels = queue.popleft()
        for j in range(k):
            if j in levels and j != bsub:
                continue
            
            for nbr, value in graph.nodeEdges[curNode][j]:
                if nbr in visited or edge(curNode, nbr) in removed or nbr < lowerBound[j] or nbr >= upperBound[j]:
                    continue
                
                visited.add(nbr)
                backPointers[nbr] = curNode                
                #stack.append(nbr)
                
                newlvls = set(levels)
                newlvls.add(j)
                queue.append((nbr, newlvls))
                
                if nbr == nodeB:
                    path = []
                    cur = nbr
                    while cur != nodeA:
                        prev = backPointers[cur]
                        path.append(edge(cur, prev))
                        cur = prev
        
                    return path        
    return None

def edge(a, b):
    return (min(a, b), max(a, b))

def getBoundPercent(bound, context):
    return int(100 * sum([bound[i] - context.lowerBound[i] for i in range(len(bound))]) / context.numNodes)

class MwtSearchContext:
    
    def __init__(self, lowerBound, upperBound):
        self.numNodes = sum([upperBound[i] - lowerBound[i] for i in range(len(lowerBound))])
        self.percentDone = 0
        self.maxFrontier = list(lowerBound)
        self.fullFrontier = list(lowerBound)
        self.lowerBound = list(lowerBound)
        self.upperBound = list(upperBound)       
        
        
class MwtSearchState:
    
    stateCounter = 0
        
    def __init__(self, otherState = None):
        self.frontier = []
        self.removed = set()
        self.frontierRemoved = set()
        self.cost = 0
        
        
        MwtSearchState.stateCounter = MwtSearchState.stateCounter + 1
        self.count = MwtSearchState.stateCounter
        
        if otherState is not None:
            self.frontier = list(otherState.frontier)
            self.removed = set(otherState.removed)
            self.frontierRemoved = set(otherState.frontierRemoved)
            self.cost = otherState.cost
    
       
    def getHeuristic(self):
        return (self.cost, -self.count)
    
    def getHeuristic2(self, lowerBound, upperBound):
        lsum = sum([self.frontier[i] - lowerBound[i] for i in range(len(self.frontier))])
        rsum = sum([upperBound[i] - self.frontier[i] for i in range(len(self.frontier))])
        #return (self.cost/max(lsum,1), -self.count)
        return (self.cost*(1 + rsum/max(lsum,1)), -self.count)
    
    def getStateKey(self):
        #return tuple(self.frontier + [self.cost, len(self.removed)])
        #return tuple(self.frontier)
        return (tuple(self.frontier), frozenset(self.frontierRemoved))
        #return (tuple(self.frontier), self.cost, len(frozenset(self.frontierRemoved)))

'''
def buildEdgeSets(graph):
    k = len(graph.subalignments)
    edgeSets = []
    for node in graph.nodeEdges:
        for i in range(k):
            if len(graph.nodeEdges[node][i]) > 1:
                edgeSets.append([edge(node, nbr) for nbr, value in graph.nodeEdges[node][i]])
                
    return edgeSets

def ilpStage(graph):
    edgeSets = buildEdgeSets(graph)
    removed, cost = edgeSetsAndCyclesToIlp(graph, edgeSets, [])
    Configs.log("Preprocessed with ILP, {} removed, {} cost..".format(len(removed), cost))
    return removed, cost


def mwtIlp(graph, lowerBound, upperBound):
    context = MwtSearchContext(lowerBound, upperBound)    
    state = MwtSearchState()
    state.frontier = list(lowerBound)       
    bestClusters, bestCost, newCycles = greedySearch(graph, state, context)
    cycles = newCycles 
    Configs.log("Initial greedy search: {} cost, {} new cycles found..".format(bestCost, len(newCycles)))
    
    iteration = 0
    while len(newCycles) > 0 and iteration < 10:
        iteration = iteration + 1 
        removed, cost = edgeSetsAndCyclesToIlp(graph, [], cycles)    
        Configs.log("ILP refinement {}: {} cost over {} cycles..".format(iteration, cost, len(cycles)))
            
        context = MwtSearchContext(lowerBound, upperBound)    
        state = MwtSearchState()
        state.frontier = list(lowerBound)   
        state.removed = removed 
        state.cost = cost
        clusters, newCost, newCycles = greedySearch(graph, state, context)
        Configs.log("Greedy search {}: {} cost, {} new cycles found..".format(iteration, newCost, len(newCycles)))
        cycles.extend(newCycles)
        
        if newCost < bestCost:
            bestCost = newCost
            bestClusters = clusters
        else:
            break
    
    return bestClusters, bestCost

def edgeSetsAndCyclesToIlp(graph, edgeSets, cycles):
    from ortools.linear_solver import pywraplp
    solver = pywraplp.Solver('SolveIntegerProblem', pywraplp.Solver.CBC_MIXED_INTEGER_PROGRAMMING)
    #solver.EnableOutput()
    objective = solver.Objective()
    
    edgeVars = {}
    for edgeSet in edgeSets:
        edgeConstraint = solver.Constraint(len(edgeSet)-1, len(edgeSet))
        for edge in edgeSet:
            if edge not in edgeVars:            
                edgeVar = solver.IntVar(0, 1, "Edge_{}".format(edge)) 
                edgeVars[edge] = edgeVar
                objective.SetCoefficient(edgeVar, graph.matrix[edge[0]][edge[1]])
            else:
                edgeVar = edgeVars[edge]
            edgeConstraint.SetCoefficient(edgeVar, 1)
    
    for cycle in cycles:
        cycleConstraint = solver.Constraint(1, len(cycle))
        for edge in cycle:
            if edge not in edgeVars:            
                edgeVar = solver.IntVar(0, 1, "Edge_{}".format(edge)) 
                edgeVars[edge] = edgeVar
                objective.SetCoefficient(edgeVar, graph.matrix[edge[0]][edge[1]])
            else:
                edgeVar = edgeVars[edge]
            cycleConstraint.SetCoefficient(edgeVar, 1)
            
    Configs.log("Number of variables: {} ".format(solver.NumVariables()))
    Configs.log("Number of constraints: {}".format(solver.NumConstraints()))
    objective.SetMinimization() 
    status = solver.Solve()
    cost = solver.Objective().Value()
    Configs.log(status)    
    Configs.log("Result: {}".format(cost))
    
    removed = []
    for edge in edgeVars:
        if edgeVars[edge].solution_value() == 1:
            removed.append(edge)
    return set(removed), cost


def edgeSetsAndCyclesToIlpAlternate(graph, edgeSets, cycles):
    from ortools.sat.python import cp_model
    model = cp_model.CpModel()
    solver = cp_model.CpSolver()
    
    edgeVars = {}
    for edgeSet in edgeSets:
        for edge in edgeSet:
            if edge not in edgeVars:
                edgeVars[edge] = model.NewBoolVar("Edge_{}".format(edge))
        model.Add(sum([edgeVars[edge] for edge in edgeSet]) >= len(edgeSet)-1)
    
    for cycle in cycles:
        for edge in cycle:
            if edge not in edgeVars:        
                edgeVars[edge] = model.NewBoolVar("Edge_{}".format(edge))
        model.Add(sum([edgeVars[edge] for edge in cycle]) >= 1)
    
    model.Minimize(sum([graph.matrix[edge[0]][edge[1]]*edgeVars[edge] for edge in edgeVars])) 
    print(model.ModelStats())
    solver.parameters.log_search_progress = True
    solver.parameters.max_time_in_seconds = 600.0
    status = solver.Solve(model)
    cost = solver.ObjectiveValue()
    print(status)    
    print("Result =", cost)  
    
    removed = []
    for edge in edgeVars:
        if solver.Value(edgeVars[edge]) == 1:
            removed.append(edge)
    return set(removed), cost
'''
    


'''
Created on Oct 26, 2020

@author: Vlad
'''

import os
import json
import time
import random
from tasks import task


def writeTasksToFile(taskList, tasksFile, append = True):
    if not append and len(taskList) == 0 and os.path.exists(tasksFile):
        os.remove(tasksFile)
    elif len(taskList) > 0:
        with open(tasksFile, 'a' if append else 'w') as file:
            for t in taskList:
                file.write(t.json + "\n")

def readTasksFromFile(tasksFile):
    fileTasks = []
    if os.path.exists(tasksFile):
        with open(tasksFile) as file:
            for line in file:
                mapper = json.loads(line.strip())
                fileTasks.append(task.Task(**mapper))
    return fileTasks


class FileLock:
    
    def __init__(self, filePath):
        self.filePath = filePath
    
    def __enter__(self):
        while True:
            try:
                lock = open(self.filePath, 'x')
                lock.close()
                return self
            except:
                time.sleep(random.random()*0.1 + 0.05)
                #time.sleep(random.random() + 0.5)
            
    def __exit__(self, excType, excVal, excTb):
        os.remove(self.filePath)



'''
Created on Aug 23, 2020

@author: Vlad
'''

import heapq
from collections import deque 

from configuration import Configs

'''
Region-growing strategy, similar to Kruskal's algorithm. Start with an empty graph, greedily
adding the heaviest edges first. Can be used for clustering or tracing.
'''

def rgSearch(graph):
    Configs.log("Finding graph trace with region-growing search..")
    
    k = len(graph.context.subalignments)
    lowerBound = [graph.subsetMatrixIdx[i] for i in range(k)]
    upperBound = [graph.subsetMatrixIdx[i] + graph.subalignmentLengths[i] for i in range(k)] 
    graph.clusters = rgCluster(graph, lowerBound, upperBound, True)

def rgCluster(graph, lowerBound, upperBound, enforceTrace = True):
    clusters = []
    clusterPointers = {}
    clusterPos = {}
    nodeClusters = {}
    weightMap = []
    absorbed = set()
    cantConnects = set()
    
    for s in range(len(lowerBound)):
        for a in range(lowerBound[s], upperBound[s]):
            clusters.append([a])
            idx = len(clusters)-1
            nodeClusters[a] = idx
            weightMap.append({})
            clusterPos[idx] = {s : a}
            clusterPointers[idx] = {s : (idx-1 if idx > lowerBound[s] else None, idx+1 if idx < upperBound[s]-1 else None)}
    
    
    heap = buildHeap(graph, nodeClusters, weightMap, lowerBound, upperBound)
    Configs.log("Built a heap of size {}..".format(len(heap)))
    crunchHeap(graph, heap, clusters, nodeClusters, clusterPos, clusterPointers, weightMap, cantConnects, absorbed, enforceTrace)

    #c2 = [sorted(c) for c in clusters if len(c) > 0]
    #c2.sort(key= lambda l : graph.matSubPosMap[l[0]])
    #for c in c2:
    #    print(c)
        
    if enforceTrace:
        clusters = orderClusters(graph, clusters, nodeClusters, lowerBound, upperBound)
        #for c in clusters:
        #    print(sorted(c))
    return clusters

def buildHeap(graph, nodeClusters, weightMap, lowerBound, upperBound):
    heap = []
    for s in range(len(lowerBound)):
        for a in range(lowerBound[s], upperBound[s]):
            asub, apos = graph.matSubPosMap[a]
            i = nodeClusters[a]
            for b, value in graph.matrix[a].items():
                bsub, bpos = graph.matSubPosMap[b]
                if b <= a or asub == bsub or b < lowerBound[bsub] or b >= upperBound[bsub]:
                    continue
                j = nodeClusters[b]
                weightMap[j][i] = value
                weightMap[i][j] = value
                heapq.heappush(heap, (-1 * value, a, b))
    
    return heap
    #baseIdx = max(range(k), key = lambda x : upperBound[x] - lowerBound[x])
    #baseLength = upperBound[baseIdx] - lowerBound[baseIdx]

def crunchHeap(graph, heap, clusters, nodeClusters, clusterPos, clusterPointers, weightMap, cantConnects, absorbed, enforceTrace):
    while len(heap) > 0:
        value, a, b = heapq.heappop(heap)
        i, j = nodeClusters[a], nodeClusters[b]
        if i == j or orderPair(i,j) in cantConnects:
            continue
        
        if not checkConnect(graph, i, j, clusters, clusterPos, enforceTrace):
            cantConnects.add(orderPair(i,j))
            continue
        
        absorbed.add(j)
        for e in clusters[j]:
            nodeClusters[e] = i
            clusters[i].append(e)
            asub, apos = graph.matSubPosMap[e]
            clusterPos[i][asub] = e
        clusters[j] = []

        if enforceTrace:        
            for s in clusterPointers[j]:
                prev, nxt = clusterPointers[j][s]
                if prev is not None:
                    clusterPointers[prev][s] = (clusterPointers[prev][s][0], i) 
                if nxt is not None:
                    clusterPointers[nxt][s] = (i, clusterPointers[nxt][s][1])
                clusterPointers[i][s] = (prev, nxt)
    
            updateMergePointers(graph, i, clusterPointers, clusters, clusterPos)
         
        #print("Clusters left: {}".format(len(clusters) - len(absorbed)))   
        for n in weightMap[j]:
            if n in absorbed:
                continue
            weightMap[i][n] = weightMap[i].get(n, 0) + weightMap[j][n]
            weightMap[n][i] = weightMap[i][n]
            heapq.heappush(heap, (-1 * weightMap[i][n], clusters[i][0], clusters[n][0]))

def updateMergePointers(graph, i, clusterPointers, clusters, clusterPos):
    subsets = [graph.matSubPosMap[a][0] for a in clusters[i]]
    
    for s in subsets:
        queue = deque([i])
        visited = set([i])
        
        while len(queue) > 0:
            curNode = queue.popleft()
            
            if clusterPos[curNode].get(s, float('inf')) > clusterPos[i][s] or curNode == i:
                clusterPos[curNode][s] = clusterPos[i][s]
                
                for p in clusterPointers[curNode]:
                    prv, nxt = clusterPointers[curNode][p]
                    if prv not in visited and prv is not None:
                        queue.append(prv)
                        visited.add(prv)
           
            
def checkConnect(graph, i, j, clusters, clusterPos, enforceTrace):
    ci , cj = set([graph.matSubPosMap[a][0] for a in clusters[i]]), set([graph.matSubPosMap[a][0] for a in clusters[j]])
    for s in ci:
        if s in cj:
            return False
    
    if not enforceTrace:
        return True
    
    for s in ci:
        if clusterPos[j].get(s, float('inf')) <= clusterPos[i][s]:
            return False
    for s in cj:
        if clusterPos[i].get(s, float('inf')) <= clusterPos[j][s]:
            return False    
    return True

def orderClusters(graph, clusters, nodeClusters, lowerBound, upperBound):
    orderedClusters = []
    frontier = list(lowerBound)
    while True:
        foundGood = False
        for j in range(len(lowerBound)):
            good = True
            idx = frontier[j]
            if idx >= upperBound[j]:
                continue
            i = nodeClusters[idx]
            for b in clusters[i]:
                bsub, bpos = graph.matSubPosMap[b]
                if b > frontier[bsub]:
                    #print(bsub, b, frontier[bsub])
                    good = False
                    break
                
            if good:
                orderedClusters.append(clusters[i])
                for b in clusters[i]:
                    bsub, bpos = graph.matSubPosMap[b]
                    frontier[bsub] = b + 1
                foundGood = True
                break
        if not foundGood:
            break
    return orderedClusters

def orderPair(a, b):
    return (min(a, b), max(a, b))


'''
Created on Aug 23, 2020

@author: Vlad
'''

import heapq
import time

from configuration import Configs

'''
Optimizer may be used to post-process a trace to improve the MWT score by shuffling nodes between clusters.
Disabled by default - tends to be very time-consuming with negligible improvements to accuracy.
However, may be helpful when using inaccurate clustering and/or tracing algorithms.
The quality of the resulting trace is usually on par with using MCL/minclusters.
'''

def optimizeTrace(graph):
    time1 = time.time() 
    
    if Configs.graphTraceOptimize:
        Configs.log("Optimization pass..")
        graph.addSingletonClusters()
        graph.clusters = optimizeClusters(graph, graph.clusters)
        Configs.log("Optimized the trace to {} clusters with a total cost of {}".format(len(graph.clusters), graph.computeClusteringCost(graph.clusters)))
    else:
        Configs.log("Skipping optimization pass..")
    time2 = time.time()
    Configs.log("Finished optimization in {} sec..".format(time2-time1))
    

def optimizeClusters(graph, clusters):
    bestClusters, bestCost = clusters, graph.computeClusteringCost(clusters)
    Configs.log("Starting optimization from initial cost of {}..".format(bestCost)) 
    context = SearchContext(clusters)
    context.initialize(graph)
    
    passNum = 1
    while True:
        Configs.log("Starting optimization pass {}..".format(passNum))
        newClusters, gain = optimizationPass(graph, bestClusters, context)
        if gain > 0:
            bestClusters = newClusters
            bestCost = bestCost - gain
            Configs.log("New clustering with a cost of {} over {} clusters..".format(bestCost, len(bestClusters)))
            #Configs.log("Verifying cost of {}..".format(graph.computeClusteringCost(bestClusters)))
        else:
            break
        passNum = passNum + 1
    #Configs.log("Final optimized cost of {} over {} clusters..".format(graph.computeClusteringCost(bestClusters), len(bestClusters)))
    Configs.log("Final optimized cost of {} over {} clusters..".format(bestCost, len(bestClusters)))
    return bestClusters

def optimizationPass(graph, clusters, context):
    context.initializeHeap(graph)
    #cost = graph.computeClusteringCost(clusters)
    #Configs.log("Initial cost: {}..".format(cost))
    
    bestGain, currentGain, bestClusters = 0, 0, clusters
    
    #move = 0
    while True:
        nextMove = getNextClusterMove(graph, context)
        if nextMove is None:
            if currentGain == bestGain:
                bestClusters = context.getOrderedClusters()
            break
            
        gain, element, src, dest, updateList = nextMove        
        if gain < 0 and currentGain == bestGain:
            bestClusters = context.getOrderedClusters()
            break
        
        currentGain = currentGain + gain
        if currentGain > bestGain:
            bestGain = currentGain
            context.locked = set()
        else:
            context.locked.add(element)
            
        context.moveElements(graph, element, dest, updateList)
        context.updateMoves(graph, element, src, dest, updateList) 
        
        #move = move + 1
        #if move % 10000 == 0:       
        #Configs.log("Current / best cost: {} / {}..".format(cost - currentGain, cost - bestGain))
    return bestClusters, bestGain

def getNextClusterMove(graph, context):
    while len(context.heap) > 0:
        gain, element, dest = heapq.heappop(context.heap)
        gain = gain * -1
        src = context.elementClusters[element]
        context.elementMoves[element, dest] = None
        
        if element in context.locked or dest in context.deletedClusters:
            continue
        
        simpleGain = context.getGainSimple(element, dest)
        if simpleGain <= 0:
            continue
        
        updateList = context.getElementUpdateList(graph, element, dest)
        updatedGain = context.getGainCorrectionFast(simpleGain, updateList)
        #Configs.log("Gain is {}, updated gain is {}, heap size {}..".format(gain, updatedGain, len(context.heap)))
        
        if updatedGain < gain:
            if updatedGain > 0:
                context.elementMoves[element, dest] = updatedGain
                heapq.heappush(context.heap, (-1*updatedGain, element, dest))
            continue
        #Configs.log("Completed the move..")
        
        return updatedGain, element, src, dest, updateList
        
    return None      


class SearchContext:
    
    def __init__(self, clusters):
        self.clusters = [set(c) for c in clusters]
        self.clusterLL = {}
        self.clusterLLHead = 0
        self.clusterOrders = {}
        self.elementClusters = {}
        self.clusterSubs = {}
        self.deletedClusters = set()
        
        self.weights = {}
        self.gainStructure = []
        self.elementMoves = {}
        self.heap = []
        self.locked = set()
        
        self.mode = "positive_moves"
        #self.mode = "adjacent_moves"
            
    def initialize(self, graph):
        Configs.log("Initializing search context data structures..")
        for i in range(len(self.clusters)):
            self.clusterOrders[i] = [i]
            self.clusterLL[i] = (i-1 if i > 0 else None, i+1 if i < len(self.clusters)-1 else None)
            for a in self.clusters[i]:
                asub, apos = graph.matSubPosMap[a]
                self.elementClusters[a] = i
                self.clusterSubs[i, asub] = a
                nbrs = self.getNeighborList(graph, a)
                self.updateNeighborWeights(None, i, nbrs)
        #Configs.log("Initialized..")
        
    def initializeHeap(self, graph):
        Configs.log("Reinitializing heap and all that stuff..")
        self.gainStructure = []
        self.elementMoves = {}
        self.heap = []
        self.locked = set()
                
        print("Working with {} clusters..".format(len(self.clusters) - len(self.deletedClusters)))
        
        k = len(graph.context.subalignments)
        self.gainStructure = [[0 for j in range(graph.subalignmentLengths[i])] for i in range(k)]        
        for i in range(k):
            for j in range(graph.subalignmentLengths[i]):
                node = graph.subsetMatrixIdx[i] + j
                weight = self.weights.get((node, self.elementClusters[node]), 0)
                self.gainStructure[i][j] = weight if j == 0 else weight + self.gainStructure[i][j-1]
        
        #if self.mode == "positive_moves":
        #    candidates = self.getPositiveMoves(graph)
        #elif self.mode == "adjacent_moves":
        #    candidates = self.getAdjacentMoves(graph)
                            
        #Configs.log("Considering {} candidate moves..".format(len(candidates)))
        #Configs.log("Choosing {} out of {} candidates..".format(limit, len(candidates)))
        #candidates = heapq.nlargest(limit, candidates)
        #for gain, i, nbr in candidates:
        #    self.pullNeighborMoves(graph, i, [(nbr, gain)])
            #gain = self.getGainSimple(nbr, i)
            #self.elementMoves[nbr, i] = gain
            #heapq.heappush(self.heap, (-1*gain, nbr, i))
        
        self.getPositiveMoves(graph)
        Configs.log("Starting with {} candidate moves..".format(len(self.heap)))
        #import sys
        #sys.exit()
    
    def getPositiveMoves(self, graph):
        k = len(graph.context.subalignments)
        candidates = []
        used = set()        
        
        clusterSubMap = [[-1 for c in self.clusters] for i in range(k)] 
        
        i = self.clusterLLHead
        while i is not None:
            prev, nxt = self.clusterLL[i]
        #for i in range(len(self.clusters)):
            #if i > 0:
            if prev is not None:
                for j in range(k):
                    clusterSubMap[j][i] = clusterSubMap[j][prev]
                    #clusterSubMap[j][i] = clusterSubMap[j][i]
            for node in self.clusters[i]:
                asub, apos = graph.matSubPosMap[node]
                clusterSubMap[asub][i] = apos
            i = nxt
        
        i = self.clusterLLHead
        while i is not None:
            prev, nxt = self.clusterLL[i]                     
        #for i in range(len(self.clusters)):
            
            for node in self.clusters[i]:
                asub, apos = graph.matSubPosMap[node]
                for nbr, value in graph.matrix[node].items():
                    bsub, bpos = graph.matSubPosMap[nbr]
                    j = self.elementClusters[nbr]
                    #if asub == bsub or j <= i:
                    if asub == bsub or self.clusterOrders[j] <= self.clusterOrders[i]:
                        continue
                    
                    if (i, nbr) not in used:
                        used.add((i, nbr))
                        gain = self.weights.get((nbr, i), 0) - self.gainStructure[bsub][bpos]
                        bound = clusterSubMap[bsub][i] - 1 if self.clusterSubs.get((i, bsub)) is not None else clusterSubMap[bsub][i]
                        if bound >= 0:
                            gain = gain + self.gainStructure[bsub][bound]
                        #if gain >= 0:
                        if gain > 0 or (gain == 0 and len(self.clusters[i]) >= len(self.clusters[j])):
                            #candidates.append((gain, i, nbr))
                            self.elementMoves[nbr, i] = gain
                            heapq.heappush(self.heap, (-1*gain, nbr, i))
                        
                    if (j, node) not in used:
                        used.add((j, node))
                        gain = self.weights.get((node, j), 0) - self.gainStructure[asub][clusterSubMap[asub][j]]
                        if apos > 0:
                            gain = gain + self.gainStructure[asub][apos-1]
                        #if gain >= 0:
                        if gain > 0 or (gain == 0 and len(self.clusters[j]) >= len(self.clusters[i])):
                            #candidates.append((gain, j, node))
                            self.elementMoves[node, j] = gain
                            heapq.heappush(self.heap, (-1*gain, node, j))             
            i = nxt
                   
        return candidates
    
    def getAdjacentMoves(self, graph):
        candidates = []
        for i in range(len(self.clusters)-1):
            for a in self.clusters[i]:
                candidates.append((0, i+1, a))
                asub, apos = graph.matSubPosMap[a]
                if a < graph.subsetMatrixIdx[asub] + graph.subalignmentLengths[asub] - 1:
                    for j in range(i, self.elementClusters[a+1]):
                        candidates.append((0, j, a+1))
                
            for a in self.clusters[i+1]:
                candidates.append((0, i, a)) 
                asub, apos = graph.matSubPosMap[a]
                if a > graph.subsetMatrixIdx[asub]:
                    for j in range(self.elementClusters[a-1]+1, i+2):
                        candidates.append((0, j, a-1))       
        return candidates
        
    def getGain(self, element, dest, updateList):
        gain = 0
        for item in updateList:
            gain = gain - self.weights.get((item, self.elementClusters[item]), 0)
        gain = gain - self.weights.get((element, self.elementClusters[element]), 0)
        gain = gain + self.weights.get((element, dest), 0)
        return gain
    
    def getGainSimple(self, element, dest):
        return self.weights.get((element, dest), 0) - self.weights.get((element, self.elementClusters[element]), 0)
    
    def getGainCorrection(self, updateList):
        gain = 0
        for item in updateList:
            gain = gain - self.weights.get((item, self.elementClusters[item]), 0)
        return gain
    
    def getGainCorrectionFast(self, gain, updateList):
        for item in updateList:
            gain = gain - self.weights.get((item, self.elementClusters[item]), 0)
            if gain < 0:
                return gain
        return gain
    
    def getElementUpdateList(self, graph, element, dest):
        asub, apos = graph.matSubPosMap[element]
        updateList = []
        
        curCluster = self.elementClusters[element]
        while curCluster != dest:
            if self.compareClusterOrder(curCluster, dest):
                prev, curCluster = self.clusterLL[curCluster]
            else:
                curCluster, nxt = self.clusterLL[curCluster]
            if self.clusterSubs.get((curCluster, asub)) is not None:
                updateList.append(self.clusterSubs[curCluster, asub])
        return updateList

    def getNeighborList(self, graph, element):
        asub, apos = graph.matSubPosMap[element]
        nbrs = []
        
        for nbr, value in graph.matrix[element].items():
            bsub, bpos = graph.matSubPosMap[nbr]
            if asub != bsub:
                nbrs.append((nbr, value))
        return nbrs
        
    def updateNeighborWeights(self, src, dest, nbrs):
        for nbr, value in nbrs:
            if src is not None:
                self.weights[nbr, src] = self.weights.get((nbr, src), 0) - value
            if dest is not None:
                self.weights[nbr, dest] = self.weights.get((nbr, dest), 0) + value
    
    def pullNeighborMoves(self, graph, dest, nbrs):
        if dest in self.deletedClusters:
            return
        for nbr, value in nbrs:
            if nbr in self.locked or self.elementClusters[nbr] == dest:
                continue               
            
            gain = self.getGainSimple(nbr, dest)
            if self.elementMoves.get((nbr, dest)) is None or gain > self.elementMoves[nbr, dest]:
                #self.elementMoves[nbr, dest] = gain
                #heapq.heappush(self.heap, (-1*gain, nbr, dest))    
                updateList = self.getElementUpdateList(graph, nbr, dest)
                #gain = gain + self.getGainCorrection(updateList)
                gain = self.getGainCorrectionFast(gain, updateList)
                if gain >= 0 and (self.elementMoves.get((nbr, dest)) is None or gain > self.elementMoves[nbr, dest]):
                #if(gain > 0 or (gain == 0 and len(self.clusters[dest]) >= len(self.clusters[self.elementClusters[nbr]]))) \
                #and (self.elementMoves.get((nbr, dest)) is None or gain > self.elementMoves[nbr, dest]):
                    self.elementMoves[nbr, dest] = gain
                    heapq.heappush(self.heap, (-1*gain, nbr, dest))
                #else:
                #    print(nbr, dest, value, gain)

    def moveElement(self, graph, element, src, dest):
        asub, apos = graph.matSubPosMap[element]
        self.clusters[src].remove(element)
        self.clusters[dest].add(element)
        self.elementClusters[element] = dest
        if self.clusterSubs[src, asub] == element:
            self.clusterSubs[src, asub] = None
        self.clusterSubs[dest, asub] = element
        if len(self.clusters[src]) == 0:# and delete:
            self.deleteCluster(src)
        nbrs = self.getNeighborList(graph, element)
        self.updateNeighborWeights(src, dest, nbrs)    
        
    def moveElements(self, graph, element, dest, updateList):
        src = self.elementClusters[element]
        curNode = dest
        self.moveElement(graph, element, src, dest)
        for node in updateList:
            prev, nxt = self.clusterLL[curNode]
            if self.clusterOrders[dest] > self.clusterOrders[src]:
                idx = self.insertCluster(curNode, nxt)
            else:
                idx = self.insertCluster(prev, curNode)
            #self.moveElement(graph, node, self.elementClusters[node], idx, self.elementClusters[node] != dest)
            self.moveElement(graph, node, self.elementClusters[node], idx)
            curNode = idx
        
    def updateMoves(self, graph, element, src, dest, updateList):
        if self.mode == "positive_moves":
            nbrs = self.getNeighborList(graph, element)
            nbrs = [item for item in nbrs if self.getGainSimple(item[0], dest) >= 0]
            self.pullNeighborMoves(graph, dest, nbrs)
            #self.pullNeighborMoves(graph, src, [(element, 0)])
            for node in updateList:
                nbrs = self.getNeighborList(graph, node)
                nbrs = [item for item in nbrs if self.getGainSimple(item[0], self.elementClusters[node]) >= 0]
                self.pullNeighborMoves(graph, self.elementClusters[node], nbrs)
        elif self.mode == "adjacent_moves":
            for a in updateList:
                nbrs = []
                asub, apos = graph.matSubPosMap[a]
                i = self.elementClusters[a]
                prv, nxt = self.clusterLL[i]
                if prv is not None:
                    for b in self.clusters[prv]:
                        nbrs.append((b, 0))
                if nxt is not None:
                    for b in self.clusters[nxt]:
                        nbrs.append((b, 0))
                if a < graph.subsetMatrixIdx[asub] + graph.subalignmentLengths[asub] - 1:
                    nbrs.append((a+1, 0))
                if a > graph.subsetMatrixIdx[asub]:
                    nbrs.append((a-1, 0))
                self.pullNeighborMoves(graph, i, nbrs)
    
    def insertCluster(self, prev, nxt):
        self.clusters.append(set())
        idx = len(self.clusters)-1
        self.clusterOrders[idx] = self.getMiddleOrder(prev, nxt)        
        self.clusterLL[idx] = (prev, nxt)
        if prev is not None:
            self.clusterLL[prev] = (self.clusterLL[prev][0], idx) 
        if nxt is not None:
            self.clusterLL[nxt] = (idx, self.clusterLL[nxt][1])
            
        if prev is None:
            self.clusterLLHead = idx    
        return idx        
    
    def deleteCluster(self, cluster):
        self.deletedClusters.add(cluster)
        prev, nxt = self.clusterLL[cluster]
        if prev is not None:
            self.clusterLL[prev] = (self.clusterLL[prev][0], nxt) 
        if nxt is not None:
            self.clusterLL[nxt] = (prev, self.clusterLL[nxt][1])
        
        if prev is None:
            self.clusterLLHead = nxt
    
    def getMiddleOrder(self, a, b):
        #return (self.clusterOrders[a] + self.clusterOrders[b]) * 0.5
        oa, ob = self.clusterOrders.get(a, []), self.clusterOrders.get(b, [])
        idx = len(oa)-1
        order = list(oa)
        if len(ob) == len(oa):
            order.append(0)
        elif len(ob) < len(oa):            
            order[-1] = order[-1] + 1
        elif len(ob) > len(oa):
            order.append(ob[idx+1]-1)
        
        #print(oa, ob, order)    
        return order
                
    
    def compareClusterOrder(self, a, b):
        oa, ob = self.clusterOrders[a], self.clusterOrders[b]
        idx = 0
        while True:            
            x, y = oa[idx] if idx < len(oa) else -float('inf'), ob[idx] if idx < len(ob) else -float('inf')
            if x < y:
                return True
            if y < x:
                return False
            idx = idx + 1
    
    def getOrderedClusters(self):
        orderedClusters = []
        cur = self.clusterLLHead
        while cur is not None:
            if len(self.clusters[cur]) > 0:
                orderedClusters.append(list(self.clusters[cur]))
            prv, cur = self.clusterLL[cur]
        #for c in orderedClusters:
        #    print(c)
        return orderedClusters


'''
Created on May 29, 2020

@author: Vlad
'''

import os
import shutil
import time
import random

from align.decompose import decomposer
from helpers import sequenceutils, hmmutils, treeutils
from tasks import task
from tools import external_tools
from configuration import Configs

'''
Different options for estimating a guide tree.
The main ones are FastTree (for accuracy) and Clustal Omega's mbed (for speed).
'''

def buildInitialTree(context, workingDir, treeType):
    if treeType is not None and os.path.exists(treeType):
        Configs.log("Found user guide tree {}".format(treeType))
        return treeType
    
    tempDir = os.path.join(workingDir, "initial_tree")
    outputTreePath = os.path.join(tempDir, "initial_tree.tre")
    if os.path.exists(outputTreePath):
        Configs.log("Found existing initial tree {}".format(outputTreePath))
        return outputTreePath
    if os.path.exists(tempDir):
        shutil.rmtree(tempDir)
    os.makedirs(tempDir)
    
    time1 = time.time() 
    
    if treeType is None or treeType.lower() == "fasttree": 
        Configs.log("Building PASTA-style FastTree initial tree on {} with skeleton size {}..".format(context.sequencesPath, Configs.decompositionSkeletonSize))
        alignPath = os.path.join(tempDir, "initial_align.txt")
        buildInitialAlignment(context.unalignedSequences, tempDir, Configs.decompositionSkeletonSize, None, alignPath)
        external_tools.runFastTree(alignPath, tempDir, outputTreePath, "fast").run()
    elif treeType is None or treeType.lower() == "fasttree-noml": 
        Configs.log("Building PASTA-style FastTree (NO ML) initial tree on {} with skeleton size {}..".format(context.sequencesPath, Configs.decompositionSkeletonSize))
        alignPath = os.path.join(tempDir, "initial_align.txt")
        buildInitialAlignment(context.unalignedSequences, tempDir, Configs.decompositionSkeletonSize, None, alignPath)
        external_tools.runFastTree(alignPath, tempDir, outputTreePath, "noml").run()
    elif treeType.lower() == "parttree":
        Configs.log("Building MAFFT PartTree initial tree on {}..".format(context.sequencesPath))
        taxa = list(context.unalignedSequences.keys())
        external_tools.runMafftGuideTree(context.sequencesPath, tempDir, outputTreePath, Configs.numCores).run()
        treeutils.convertMafftGuideTree(outputTreePath, taxa)
    elif treeType.lower() == "clustal":
        Configs.log("Building Clustal Omega initial tree on {}..".format(context.sequencesPath))
        external_tools.runClustalOmegaGuideTree(context.sequencesPath, tempDir, outputTreePath, Configs.numCores).run()
    else:
        raise Exception("Guide tree {} not a file and not recognized..".format(treeType))

    time2 = time.time()
    Configs.log("Built initial tree on {} in {} sec..".format(context.sequencesPath, time2-time1))
    
    return outputTreePath

def buildInitialAlignment(sequences, tempDir, skeletonSize, initialAlignSize, outputAlignPath):
    skeletonPath = os.path.join(tempDir, "skeleton_sequences.txt")
    queriesPath = os.path.join(tempDir, "queries.txt") 
    hmmDir = os.path.join(tempDir, "skeleton_hmm")
    hmmPath = os.path.join(hmmDir, "hmm_model.txt")
    initialInsertPath = os.path.join(tempDir, "initial_insert_align.txt")
    if not os.path.exists(hmmDir):
        os.makedirs(hmmDir)
    
    if initialAlignSize is None or initialAlignSize > len(sequences):
        initialAlignSize = len(sequences)
    
    skeletonTaxa, remainingTaxa = decomposer.chooseSkeletonTaxa(sequences, skeletonSize)
    additional = initialAlignSize-skeletonSize
    random.shuffle(remainingTaxa)
    remainingTaxa, unusedTaxa = remainingTaxa[:additional], remainingTaxa[additional:]
    
    sequenceutils.writeFasta(sequences, skeletonPath, skeletonTaxa)
    external_tools.runMafft(skeletonPath, None, tempDir, outputAlignPath, Configs.numCores).run()
    
    if len(remainingTaxa) > 0:
        sequenceutils.writeFasta(sequences, queriesPath, remainingTaxa)    
        hmmutils.buildHmmOverAlignment(outputAlignPath, hmmPath).run()
        hmmTasks = hmmutils.hmmAlignQueries(hmmPath, queriesPath)
        task.submitTasks(hmmTasks)
        for hmmTask in task.asCompleted(hmmTasks):
            hmmutils.mergeHmmAlignments([hmmTask.outputFile], outputAlignPath, includeInsertions=False)
            if Configs.graphBuildMethod == "initial":
                hmmutils.mergeHmmAlignments([hmmTask.outputFile], initialInsertPath, includeInsertions=True)



'''
Created on Aug 23, 2020

@author: Vlad
'''

import heapq
from collections import deque 

from configuration import Configs

def rgFastSearch(graph):
    Configs.log("Finding graph trace with fast region-growing search..")
    
    k = len(graph.context.subalignments)
    lowerBound = [graph.subsetMatrixIdx[i] for i in range(k)]
    upperBound = [graph.subsetMatrixIdx[i] + graph.subalignmentLengths[i] for i in range(k)] 
    cuts = rgFastCluster(graph, lowerBound, upperBound, True)
    graph.clusters = cutsToClusters(graph, cuts)

def rgFastCluster(graph, lowerBound, upperBound, enforceTrace = True):
    initialCuts = initialSplit(graph, lowerBound, upperBound, enforceTrace)
    if len(initialCuts) == 2:
        return initialCuts
    #Configs.log("Starting with {} coarse cuts..".format(len(initialCuts)))
    cuts = []
    for i in range(len(initialCuts)-1):
        intervalCuts = rgFastCluster(graph, initialCuts[i], initialCuts[i+1], enforceTrace)
        cuts.extend(intervalCuts[:-1])
    
    cuts.append(list(upperBound))
    #Configs.log("Returning {} fine cuts..".format(len(cuts)))
    return cuts
        
def initialSplit(graph, lowerBound, upperBound, enforceTrace = True):
    k = len(graph.context.subalignments)
    baseIdx = max(range(k), key = lambda x : upperBound[x] - lowerBound[x])
    #baseIdx = min(range(k), key = lambda x : upperBound[x] - lowerBound[x] if upperBound[x] - lowerBound[x] >= 2 else float('inf'))
    baseLength = upperBound[baseIdx] - lowerBound[baseIdx]
    if baseLength < 2:
        return [list(lowerBound), list(upperBound)]
    
    clusters = initialSplitExpansion(graph, lowerBound, upperBound, baseIdx, baseLength)
    #clusters = initialSplitExpansionSimple(graph, lowerBound, upperBound, baseIdx, baseLength)
    
    cuts = clustersToCuts(graph, lowerBound, upperBound, clusters)   
        
    return cuts

def initialSplitExpansion(graph, lowerBound, upperBound, baseIdx, baseLength):
    k = len(graph.context.subalignments)
    clusters = [[lowerBound[baseIdx] + i] for i in range(baseLength)]
    #idxSets = [set([baseIdx]) for i in range(baseLength)]
    idxSets = {(i, baseIdx) : lowerBound[baseIdx] + i for i in range(baseLength)}
    usedNodes = set()
    weightMap = {}
    
    boundsMap = {}
    for i in range(k):
        boundsMap[0, i] = (lowerBound[i]-1, upperBound[i])
        boundsMap[baseLength-1, i] = (lowerBound[i]-1, upperBound[i])            
    
    heap = []
    for node in range(lowerBound[baseIdx], upperBound[baseIdx]):
        for nbr, value in graph.matrix[node].items():
            i, pos = graph.matSubPosMap[nbr]
    #for i in range(k):
    #    for nbr, value in graph.nodeEdges[node][i]:
            
            if nbr < lowerBound[i]:
                continue
            if nbr >= upperBound[i]:
                continue
                #break
            idx = node - lowerBound[baseIdx]
            if (idx, i) in idxSets:
                continue
            
            heapq.heappush(heap, (-1*value, node, nbr, idx)) 
            weightMap[idx, nbr] = value
    
    while len(heap) > 0:
        value, a, b, idx = heapq.heappop(heap)
        if b in usedNodes:
            continue
        #asub, apos = graph.matSubPosMap[a]
        bsub, bpos = graph.matSubPosMap[b]

        if (idx, bsub) in idxSets:
            continue
        lower, upper = getBounds(boundsMap, baseLength, idx, bsub)
        if not (b > lower and b < upper):
            continue
        
        addBounds(graph, boundsMap, baseLength, idx, b)
        clusters[idx].append(b)
        idxSets[idx, bsub] = b
        usedNodes.add(b)
        
        for nbr, value in graph.matrix[b].items():
            i, pos = graph.matSubPosMap[nbr]
    #for i in range(k):
            if (idx, i) in idxSets:
                continue
            lower, upper = getBounds(boundsMap, baseLength, idx, i)
        #for nbr, value in graph.nodeEdges[b][i]:
            if nbr in usedNodes:
                continue               
             
            if nbr <= lower:
                continue
            if nbr >= upper:
                #break
                continue
            
            #print(weightMap.get((idx, nbr), 0))
            weight = value + weightMap.get((idx, nbr), 0)
            weightMap[idx, nbr] = weight
            heapq.heappush(heap, (-1*weight, b, nbr, idx))    

    return clusters


def getBounds(boundsMap, baseLength, idx, asub):
    a, b = 0, baseLength - 1
    if idx == a:
        return boundsMap[a, asub]
    if idx == b:
        return boundsMap[b, asub]
    
    midpoint = int((a+b)*0.5)
    while (midpoint, asub) in boundsMap:
        if idx == midpoint:
            return boundsMap[midpoint, asub]
        elif idx > midpoint:
            a = midpoint
        elif idx < midpoint:
            b = midpoint
        midpoint = int((a+b)*0.5)
    la, ua = boundsMap[a, asub]
    lb, ub = boundsMap[b, asub]
    return (la, ub)

def addBounds(graph, boundsMap, baseLength, idx, node):
    asub, apos = graph.matSubPosMap[node]
    a, b = 0, baseLength - 1
    
    while True:
        la, ua = boundsMap[a, asub]
        lb, ub = boundsMap[b, asub]
        if idx == a:
            boundsMap[a, asub] = (node, node)
            return
        elif node < ua:
            boundsMap[a, asub] = (la, node)
            
        if idx == b:
            boundsMap[b, asub] = (node, node)
            return
        elif node > lb:
            boundsMap[b, asub] = (node, ub)    
                
        midpoint = int((a+b)*0.5)
        if idx == midpoint:
            boundsMap[midpoint, asub] = (node, node)
            return
        elif (midpoint, asub) not in boundsMap:
            boundsMap[midpoint, asub] = (la, ub)
            
        if idx > midpoint:
            a = midpoint
        elif idx < midpoint:
            b = midpoint 
    

def initialSplitExpansionSimple(graph, lowerBound, upperBound, baseIdx, baseLength):
    k = len(graph.context.subalignments)
    clusters = [[lowerBound[baseIdx] + i] for i in range(baseLength)]
    #idxSets = [set([baseIdx]) for i in range(baseLength)]
    idxSets = {(i, baseIdx) : lowerBound[baseIdx] + i for i in range(baseLength)}
    usedNodes = set()
    weightMap = {}
    
    heap = []
    for node in range(lowerBound[baseIdx], upperBound[baseIdx]):
        for i in range(k):
            for nbr, value in graph.nodeEdges[node][i]:
                if nbr < lowerBound[i]:
                    continue
                if nbr >= upperBound[i]:
                    break
                idx = node - lowerBound[baseIdx]
                heapq.heappush(heap, (-1*value, node, nbr, idx)) 
                weightMap[idx, nbr] = value
    
    while len(heap) > 0:
        value, a, b, idx = heapq.heappop(heap)
        if b in usedNodes:
            continue
        #asub, apos = graph.matSubPosMap[a]
        bsub, bpos = graph.matSubPosMap[b]
        #if bsub in idxSets[idx]:
        #    continue
        if (idx, bsub) in idxSets or idxSets.get((idx-1, bsub), 0) > b or idxSets.get((idx+1, bsub), upperBound[bsub]) < b:
            continue
        
        
        clusters[idx].append(b)
        #idxSets[idx].add(bsub)
        idxSets[idx, bsub] = b
        usedNodes.add(b)
        
        for i in range(k):
            if (idx, i) in idxSets:
                continue
            for nbr, value in graph.nodeEdges[b][i]:
                if nbr in usedNodes:
                    continue                
                if nbr <  idxSets.get((idx-1, i), lowerBound[i]):
                    continue
                if nbr >= idxSets.get((idx+1, i), upperBound[i]):
                    break
                
                #print(weightMap.get((idx, nbr), 0))
                weight = value + weightMap.get((idx, nbr), 0)
                #weight = value
                weightMap[idx, nbr] = weight
                heapq.heappush(heap, (-1*weight, b, nbr, idx))    

    return clusters

def clustersToCuts(graph, lowerBound, upperBound, clusters):
    cuts = [list(lowerBound)]
    #cuts = []
    cut = list(lowerBound)
    for i, cluster in enumerate(clusters):
        if i == 0:
            continue
        #cut = list(cuts[-1])
        
        for a in cluster:
            asub, apos = graph.matSubPosMap[a]
            cut[asub] = max(a, cut[asub])
        cuts.append(cut)
        cut = list(cut)    
    cuts.append(list(upperBound))
    return cuts

def cutsToClusters(graph, cuts):
    clusters = []
    for i in range(len(cuts)-1):
        cluster = []
        for j in range(len(cuts[i])):
            cluster.extend(list(range(cuts[i][j], cuts[i+1][j])))
        clusters.append(cluster)
        #print(cluster)
        
    return clusters


'''
Created on Aug 23, 2020

@author: Vlad
'''

from configuration import Configs

def purgeDuplicateClusters(graph):
    uniqueClusters = set()
    newclusters = []
    for cluster in graph.clusters:
        cluster.sort()
        clusterTuple = tuple(cluster)
        if clusterTuple not in uniqueClusters:
            uniqueClusters.add(clusterTuple)
            newclusters.append(cluster)
    graph.clusters = newclusters
    Configs.log("Purged duplicate clusters. Found {} unique clusters..".format(len(graph.clusters)))

def purgeClusterViolations(graph):
    redundantCols = {}
    redundantRows = {}
    elementScores = {}
    for a, cluster in enumerate(graph.clusters):
        for b in cluster:
            bsub, bpos = graph.matSubPosMap[b] 
            redundantCols[a, bsub] = redundantCols.get((a, bsub), []) + [(a, b)] 
            redundantRows[b] = redundantRows.get(b, []) + [(a, b)]
            
            scoresum = 0
            for c in cluster:
                csub, cpos = graph.matSubPosMap[c]
                if bsub != csub:
                    scoresum  = scoresum  + graph.matrix[b].get(c,0)
            elementScores[a, b] = scoresum 
    
    problemCols = [(a,b) for a,b in redundantCols if len(redundantCols[a,b]) > 1]
    problemRows = [a for a in redundantRows if len(redundantRows[a]) > 1]
    Configs.log("Found {} row violations and {} column violations..".format(len(problemRows), len(problemCols)))
    
    sortedScores = list(elementScores.keys())
    sortedScores.sort(key = lambda x : elementScores[x])
    
    for a,b in sortedScores:
        bsub, bpos = graph.matSubPosMap[b] 
        if len(redundantCols[a, bsub]) > 1 or len(redundantRows[b]) > 1:
            graph.clusters[a].remove(b)
            redundantCols[a, bsub].remove((a,b))
            redundantRows[b].remove((a,b))                
    
    problemCols = [(a,b) for a,b in redundantCols if len(redundantCols[a,b]) > 1]
    problemRows = [a for a in redundantRows if len(redundantRows[a]) > 1]
    Configs.log("Finished violations sweep. Now {} row violations and {} column violations..".format(len(problemRows), len(problemCols)))
    
    graph.clusters = [cluster for cluster in graph.clusters if len(cluster) > 1]
    Configs.log("Purged cluster violations. Found {} clean clusters..".format(len(graph.clusters)))
    


'''
Created on Oct 26, 2018

@author: Vlad
'''

import dendropy
from dendropy.utility import bitprocessing
from helpers import sequenceutils
from configuration import Configs

import os
import sys
import random


def loadTree(treePath, nameSpace=None):
    tree = dendropy.Tree.get(path=treePath, schema="newick", preserve_underscores=True)
    if nameSpace is None:
        nameSpace = tree.taxon_namespace
    else:
        tree.migrate_taxon_namespace(nameSpace)

    tree.is_rooted = False
    tree.resolve_polytomies(limit=2)
    tree.collapse_basal_bifurcation()
    tree.update_bipartitions()
    return tree

def writeTree(tree, outputPath):
    resultTreeString = tree.as_string(schema="newick", suppress_rooting = True)
    with open(outputPath, "w") as f:
        f.write(resultTreeString) 

def compareTreesFromDendropy(tr1, tr2):
    return compareDendropyTrees(dendropy.Tree(tr1), dendropy.Tree(tr2))

#courtesy Erin Molloy
def compareTreesFromPath(treePath1, treePath2):
    print("Comparing {} with {}".format(treePath1, treePath2))
    
    tax = dendropy.TaxonNamespace()
    tr1 = dendropy.Tree.get(path=treePath1,
                            schema='newick',
                            rooting='force-unrooted',
                            taxon_namespace=tax,
                            preserve_underscores=True)
    tr2 = dendropy.Tree.get(path=treePath2,
                            schema='newick',
                            rooting='force-unrooted',
                            taxon_namespace=tax,
                            preserve_underscores=True)

    tr1.collapse_basal_bifurcation(set_as_unrooted_tree=True)
    tr2.collapse_basal_bifurcation(set_as_unrooted_tree=True)

    return compareDendropyTrees(tr1, tr2)
    #print("RF distance on %d shared leaves: %d" % (nl, fp + fn))

#courtesy Erin Molloy
def compareDendropyTrees(tr1, tr2):
    from dendropy.calculate.treecompare \
        import false_positives_and_negatives

    lb1 = set([l.taxon.label for l in tr1.leaf_nodes()])
    lb2 = set([l.taxon.label for l in tr2.leaf_nodes()])
    
    print("Comparing trees with {}|{} and {}|{} leaves..".format(len(lb1), len(tr1.leaf_nodes()), len(lb2), len(tr2.leaf_nodes())))
    
    com = lb1.intersection(lb2)
    if com != lb1 or com != lb2:
        com = list(com)
        tns = dendropy.TaxonNamespace(com)

        tr1.retain_taxa_with_labels(com)
        tr1.migrate_taxon_namespace(tns)

        tr2.retain_taxa_with_labels(com)
        tr2.migrate_taxon_namespace(tns)
    com = list(com)

    tr1.update_bipartitions()
    tr2.update_bipartitions()

    nl = len(com)
    ei1 = len(tr1.internal_edges(exclude_seed_edge=True))
    ei2 = len(tr2.internal_edges(exclude_seed_edge=True))

    [fp, fn] = false_positives_and_negatives(tr1, tr2)
    rf = float(fp + fn) / (ei1 + ei2)

    return (nl, ei1, ei2, fp, fn, rf)

def decomposeGuideTree(subsetsDir, sequencesPath, guideTreePath, maxSubsetSize, maxNumSubsets):
    sequences = sequenceutils.readFromFasta(sequencesPath, removeDashes = False)
    guideTree = dendropy.Tree.get(path=guideTreePath, schema="newick", preserve_underscores=True)
    guideTree.collapse_basal_bifurcation()
    
    for edge in guideTree.postorder_edge_iter():
        if len(edge.head_node.child_edges()) > 0:
            edge.childs = sum([e.childs for e in edge.head_node.child_edges()])
        else:
            edge.childs = 1
    guideTree.childs = guideTree.seed_node.edge.childs       
    trees = decomposeTree(guideTree, maxSubsetSize, maxNumSubsets)
    
    taxonSubsets = []
    for tree in trees:
        keep = [n.taxon.label for n in tree.leaf_nodes()]
        taxonSubsets.append(keep)
    
    subsetPaths = []
    for n, subset in enumerate(taxonSubsets):
        subsetPath = os.path.join(subsetsDir, "subset_{}.txt".format(n+1))
        subsetPaths.append(subsetPath)                    
        sequenceutils.writeFasta(sequences, subsetPath, subset) 
    return subsetPaths

def decomposeTree(tree, maxSubsetSize, numSubsets):
    trees = [tree]
    while len(trees) < numSubsets:
        largestTree = max(trees, key=lambda t : t.childs)
        
        if maxSubsetSize is not None and largestTree.childs <= maxSubsetSize:
            return trees
        else:
            numChilds = largestTree.childs
            e = getCentroidEdge(largestTree)
            t1, t2 = bipartitionByEdge(largestTree, e)
            Configs.log("Decomposing a tree with {} leaves into {} and {}..".format(numChilds, t1.childs, t2.childs))
            trees.remove(largestTree)
            trees = trees + [t1, t2]
    return trees

def bipartitionByEdge(tree, edge):
    newRoot = edge.head_node
    tail = edge.tail_node
    tail.remove_child(newRoot)
    newTree = dendropy.Tree(seed_node=newRoot, taxon_namespace = tree.taxon_namespace)
    newTree.childs = edge.childs
    tree.childs = tree.childs - newTree.childs
    
    curNode = tail
    while curNode is not None:
        curNode.edge.childs = curNode.edge.childs - edge.childs
        curNode = curNode.edge.tail_node
        
    newTree.collapse_basal_bifurcation()
    if tail == tree.seed_node:
        tree.collapse_basal_bifurcation()
    elif len(tail.child_nodes()) == 1:
        childs = tail.child_nodes()
        for child in childs:
            tail.remove_child(child)
            tail.parent_node.add_child(child)
        tail.parent_node.remove_child(tail)
    
    return tree, newTree

def getCentroidEdge(tree):
    bestBalance = float('inf')
    for edge in tree.postorder_edge_iter():
        if edge.tail_node is None:
            continue
        balance = abs(tree.childs/2 - edge.childs)
        if balance < bestBalance:
            bestBalance = balance
            bestEdge = edge  
    return bestEdge

def convertMafftGuideTree(treePath, taxa): 
    treeString = ""
    with open(treePath) as f:
        for line in f:
            line = line.strip()
            treeString = treeString + line
    if not treeString.endswith(";"):
        treeString = treeString + ";"    
    guideTree = dendropy.Tree.get(data=treeString, schema="newick", preserve_underscores=True)
    
    for node in guideTree.leaf_nodes():
        node.taxon.label = taxa[int(node.taxon.label)-1]
    writeTree(guideTree, treePath)

'''
def decomposeGuideTree(subsetsDir, sequencesPath, guideTreePath, maxSubsetSize, maxNumSubsets):
    sequences = sequenceutils.readFromFasta(sequencesPath, removeDashes = False)
    guideTree = loadTree(guideTreePath)
    trees = decomposeTree(guideTree, maxSubsetSize, maxNumSubsets)
    
    taxonSubsets = []
    for tree in trees:
        keep = [n.taxon.label for n in tree.leaf_nodes()]
        taxonSubsets.append(keep)
    
    subsetPaths = []
    for n, subset in enumerate(taxonSubsets):
        subsetPath = os.path.join(subsetsDir, "subset_{}.txt".format(n+1))
        subsetPaths.append(subsetPath)                    
        sequenceutils.writeFasta(sequences, subsetPath, subset) 
    return subsetPaths

def decomposeTree(tree, maxSubsetSize, numSubsets):
    trees = [tree]
    while len(trees) < numSubsets:
        largestTree = max(trees, key=lambda t : len(t.leaf_nodes()))
        
        if maxSubsetSize is not None and len(largestTree.leaf_nodes()) <= maxSubsetSize:
            return trees
        else:
            e = getCentroidEdge(largestTree)
            t1, t2 = bipartitionByEdge(largestTree, e)
            trees.remove(largestTree)
            trees = trees + [t1, t2]
    return trees

def bipartitionByEdge(tree, edge):
    newRoot = edge.head_node
    edge.tail_node.remove_child(newRoot)
    newTree = dendropy.Tree(seed_node=newRoot, taxon_namespace = tree.taxon_namespace)
    tree.update_bipartitions()
    newTree.update_bipartitions()
    return tree, newTree

def getCentroidEdge(tree):
    numLeaves = bitprocessing.num_set_bits(tree.seed_node.tree_leafset_bitmask)
    bestBalance = float('inf')
    for edge in tree.postorder_edge_iter():
        if edge.tail_node is None:
            continue
        balance = abs(numLeaves/2 - bitprocessing.num_set_bits(edge.bipartition.leafset_bitmask))
        if balance < bestBalance:
            bestBalance = balance
            bestEdge = edge    
    return bestEdge

def decomposeTreeNumSubsets(tree, numSubsets):
    trees = [tree]
    while len(trees) < numSubsets:
        trees.sort(key=lambda t : len(t.leaf_nodes()), reverse=True)
        e = getCentroidEdge(trees[0])
        t1, t2 = bipartitionByEdge(trees[0], e)
        trees = trees[1:] + [t1, t2]
    return trees

def decomposeTreeMaxSubsetSize(tree, maxSubsetSize, mode = "centroid"):
    numLeaves = len(tree.leaf_nodes())
    if numLeaves > maxSubsetSize:
        if mode == "centroid":
            e = getCentroidEdge(tree)
        elif mode == "random":
            e = getCentroidEdgeRandom(tree, maxSubsetSize/3)

        t1, t2 = bipartitionByEdge(tree, e)
        return decomposeTreeMaxSubsetSize(t1, maxSubsetSize, mode) + decomposeTreeMaxSubsetSize(t2, maxSubsetSize, mode)
    else:
        return [tree]

def getCentroidEdgeRandom(tree, minBound = 5):
    fullMask = tree.seed_node.tree_leafset_bitmask
    numLeaves = bitprocessing.num_set_bits(fullMask)
    candidates = []
    for edge in tree.postorder_internal_edge_iter():
        if edge.tail_node is None:
            continue
        
        mask = edge.bipartition.leafset_bitmask
        numMask1 = bitprocessing.num_set_bits(mask)
        numMask2 = numLeaves - numMask1
        
        if numMask1 >= minBound and numMask2 >= minBound:
            candidates.append(edge)    
            
    return random.choice(candidates)
'''



'''
Created on Apr 13, 2020

@author: Vlad
'''

import os

from configuration import Configs
from tools import external_tools


def runMlrMclClustering(graph):  
    Configs.log("Running MLR-MCL alignment graph clustering..")
    graphPath = os.path.join(graph.workingDir, "graph_mlr_mcl.txt")
    clusterPath = os.path.join(graph.workingDir, "clusters_mlr_mcl.txt")
    
    if not os.path.exists(clusterPath):
        if not os.path.exists(graphPath):
            writeGraphToFile(graph, graphPath)
        external_tools.runMlrMcl(graphPath, 30000, 0.5, 4, graph.workingDir, clusterPath).run()

    graph.clusters = readClustersFromFile(clusterPath)
    graph.writeClustersToFile(graph.clusterPath)
    
def writeGraphToFile(graph, filePath):
    Configs.log("Writing MLR-MCL graph file to {}".format(filePath))
    vertices, edges = 0, 0
    lines = []
    for i in range(len(graph.matrix)):
        pairs = graph.matrix[i].items()
        vertices = vertices + 1        
        edges = edges + len(pairs)
        lines.append(" ".join(["{} {}".format(a+1, b) for a, b in pairs]))      
            
    with open(filePath, 'w') as textFile:
        textFile.write("{} {} 1\n".format(vertices, int(edges/2)))
        for line in lines:
            textFile.write(line + "\n")
            
    Configs.log("Wrote graph with {} vertices and {} edges to {}".format(vertices, int(edges/2), filePath))

def readClustersFromFile(filePath):
    assignments = {}
    with open(filePath) as f:
        num = 0
        for line in f:
            cluster = int(line.strip())
            if cluster not in assignments:
                assignments[cluster] = [num]
            else:
                assignments[cluster].append(num) 
            num = num + 1
    clusters = [assignments[c] for c in range(len(assignments))]
    Configs.log("Found {} clusters..".format(len(clusters)))
    return clusters


import setuptools
from setuptools import find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="magus-msa",
    version="0.1.0a6",
    author="vlasmirnov",
    description="Multiple Sequence Alignment using Graph Clustering",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vlasmirnov/MAGUS",
    project_urls={
        "Bug Tracker": "https://github.com/vlasmirnov/MAGUS/issues",
    },
    entry_points={
        'console_scripts': [
            'magus=magus:main',
        ],
    },
    py_modules=['magus', 'configuration'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    package_dir={"": "."},
    packages=find_packages(),
    install_requires=["dendropy>=4.5.2"],
    python_requires=">=3.6",
)


'''
Created on Apr 14, 2020

@author: Vlad
'''

from configuration import Configs
from tools import external_tools


def runMclClustering(graph):  
    Configs.log("Running MCL alignment graph clustering..")
    external_tools.runMcl(graph.graphPath, Configs.mclInflationFactor, graph.workingDir, graph.clusterPath).run()
    graph.readClustersFromFile(graph.clusterPath)
    
    
